from flask import Blueprint

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow

alexa_deployment_blueprint = Blueprint("auth", __name__, template_folder="templates")

"""
@alexa_deployment_blueprint.route("/<account_username>/<project_url>/deployment/alexa/skill-manifest/set", methods=["POST"])
def set_alexa_skill_manifest(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_set_alexa_skill_manifest, to_json=True,
    )

def _set_alexa_skill_manifest(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    try:
        say_text_action = SayTextAction(**request.get_json())
        print(say_text_action)
    return {"success": True}
"""



from dataclasses import field
from typing import Optional
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

from inoft_vocal_engine.models.actions import SayTextAction


@dataclass
class MessageItem:
    id: str
    speech_items: Optional[list] = None
    is_callable: Optional[bool] = False
    variable_name: str = None

    def __post_init_post_parse__(self):
        if isinstance(self.id, str):
            self.variable_name = "".join([char.capitalize() for char in self.id.replace("-", "_")])
        else:
            self.variable_name = None

class SayActionCodeGen(SayTextAction):
    is_callable: bool = False
    text_to_say_if_message_item_missing: Optional[str] = "A message element is missing."
    code: str = None
    extra_args: dict = dict()

class SetVariableAction:
    def __init__(self):
        pass


import discord
#from discord.ext import commands
from discord.ext.commands import Context


def emoji_array(e_type, number):
    numeric_emojis = ['1️⃣', '2️⃣', '3️⃣', '4️⃣', '5️⃣', '6️⃣', '7️⃣', '8️⃣', '9️⃣', '🔟']
    behavior_emojis = []

    if e_type == 'num':
        tab = numeric_emojis[0:number]
    elif e_type == 'behav':
        tab = behavior_emojis[0:number]

    return tab


# general function to send polls
async def send_poll(message_to_send, destination, e_type, m_type, number):
    message = await destination.send(message_to_send)
    reactions = emoji_array(e_type, number)
    for emoji in reactions:
        await message.add_reaction(emoji)


# general function to send messages
async def send_message(message_to_send: str, destination_context: Context):
    await destination_context.send(message_to_send)


# function with all response options to aa reaction add
async def emojis_reaction_response(m_type, reaction, user):
    if m_type == 'launch':
        if reaction.emoji == '1️⃣':
            await user.send('Continuons ici !')
            await reaction.message.delete()
        elif reaction.emoji == '2️⃣':
            await reaction.message.channel.send('Continuons ici !')
            await reaction.message.delete()
        else:
            await reaction.message.channel.send('Reaction innapropriée')
    elif m_type == 'game':
        return m_type



import os
import time
from typing import List

import click

from inoft_vocal_framework.cloud_providers.aws.aws_core import AwsCore
from inoft_vocal_framework.cli.cli_cache import CliCache
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.skill_settings import skill_settings
from inoft_vocal_framework.utils.general import get_all_files_in_dir, get_all_dirs_in_dir
from inoft_vocal_framework.cli.botpress.content_element_object import ContentElement


class BotpressCore(AwsCore):
    def __init__(self):
        super().__init__(clients_to_load=[AwsCore.CLIENT_BOTO_SESSION, AwsCore.CLIENT_S3])
        self.settings = skill_settings.prompt_get_settings()
        self.selected_bot_dirname = None
        self.selected_bot_dirpath = None

    def generate_audio_contents_from_texts(self):
        raise Exception(f"generate_audio_contents_from_texts is temporarily being deprecated since when the "
                        f"inoft_vocal_framework and the inoft_vocal_engine have been splitted.")
        # todo: should we decide to re-implement the Polly integration inside the framework, or reserve it to the engine ?

        def prompt_user_to_select_botpress_dirpath():
            return click.prompt(text="What is the installation folderpath of your Botpress ?\n"
                                     "For example C:/Program Files (x86)/botpress-v12_8_1-win-x64")

        botpress_dirpath = CliCache.cache().get("botpress_dirpath").to_str(default=None)
        if botpress_dirpath is None:
            botpress_dirpath = prompt_user_to_select_botpress_dirpath()

        while not os.path.exists(botpress_dirpath):
            click.echo(f"The folderpath : {botpress_dirpath} was not found.")
            botpress_dirpath = prompt_user_to_select_botpress_dirpath()

        filenames_in_botpress_folder = get_all_files_in_dir(botpress_dirpath)
        while "bp.exe" not in filenames_in_botpress_folder:
            filenames_in_botpress_folder = get_all_files_in_dir(botpress_dirpath)
            click.echo(f"The folder you specified has your botpress folder ({botpress_dirpath} is valid.\n"
                       f"Yet the bp.exe file has not been found in your Botpress folder."
                       f"You must have not select the right folder, or you have an issue with Botpress and need to re-install it.")
            botpress_dirpath = prompt_user_to_select_botpress_dirpath()

        CliCache().cache().put("botpress_dirpath", botpress_dirpath)
        CliCache().save_cache_to_yaml()

        botpress_bots_dirpath = os.path.join(botpress_dirpath, "data", "bots")
        if not os.path.exists(botpress_bots_dirpath):
            raise Exception(f"No data/bots folder has been found in your Botpress directory. Did you already created at least one Bot ?\n"
                            f"The Inoft Vocal Engine has check at the following path : {botpress_bots_dirpath}")
        else:
            bots_dirnames = get_all_dirs_in_dir(botpress_bots_dirpath)
            if len(bots_dirnames) > 0:
                click.echo("Which Bot would like to select ?")
                for i_bot, bot_folder in enumerate(bots_dirnames):
                    click.echo(f"{bot_folder} - {i_bot + 1}")

                while True:
                    bot_number = str(click.prompt(f"Please type a number between 1 and {len(bots_dirnames)}"))
                    if bot_number.isdigit():
                        bot_index = int(bot_number) - 1
                        self.selected_bot_dirname = bots_dirnames[bot_index]
                        self.selected_bot_dirpath = os.path.join(botpress_bots_dirpath, self.selected_bot_dirname)
                        break

        self._content_elements_to_audacity(content_elements=self._text_elements_to_content_elements())

    def _text_elements_to_content_elements(self) -> List[ContentElement]:
        builtin_text_filepath = os.path.join(self.selected_bot_dirpath, "content-elements", "builtin_text.json")
        if not os.path.isfile(builtin_text_filepath):
            raise Exception(f"No file has been found at {builtin_text_filepath}")
        else:
            from inoft_vocal_framework.utils.general import load_json
            list_all_text_elements = load_json(builtin_text_filepath)

            from inoft_vocal_framework.inoft_vocal_markup.deserializer import Deserializer
            inoft_vocal_markup_deserializer = Deserializer(characters_names=["Léo", "Willie", "Menu"])

            content_elements: List[ContentElement] = list()
            for text_element in list_all_text_elements:
                current_content_element = ContentElement()

                text_element_safedict = SafeDict(text_element)
                text_content = text_element_safedict.get("formData").get("text$fr").to_str(default=None)
                if text_content is not None:
                    current_content_element.dialogues_lines = inoft_vocal_markup_deserializer.deserialize(text=text_content)
                    # todo: give the ability to select the language to deserialize

                current_content_element.id = text_element_safedict.get("id").to_str(default=None)
                current_content_element.created_by = text_element_safedict.get("createdBy").to_str(default=None)
                current_content_element.created_on = text_element_safedict.get("createdOn").to_str(default=None)
                current_content_element.modified_on = text_element_safedict.get("modifiedOn").to_str(default=None)
                content_elements.append(current_content_element)

            return content_elements

    @staticmethod
    def prompt_get_root_dirpath_to_save_audio() -> str:
        user_need_to_select_need_root_dirpath = False
        root_dirpath_to_save_audio_project = CliCache().cache().get("root_dirpath_to_save_audio_project").to_str(default=None)
        if root_dirpath_to_save_audio_project is not None:
            if not click.confirm(f"Do you want to save your audio files and folder to the following dirpath : {root_dirpath_to_save_audio_project}"):
                user_need_to_select_need_root_dirpath = True
        else:
            user_need_to_select_need_root_dirpath = True

        if user_need_to_select_need_root_dirpath is True:
            def prompt_user_to_select_root_dirpath_to_save_audio_project():
                return click.prompt(text="To which dirpath to you wish to save all of the files and folders that will contain your audio ?", type=str)

            root_dirpath_to_save_audio_project = prompt_user_to_select_root_dirpath_to_save_audio_project()
            while not os.path.isdir(root_dirpath_to_save_audio_project):
                if click.confirm(f"The dirpath {root_dirpath_to_save_audio_project} did not exist. Do you want to create it ?"):
                    os.makedirs(root_dirpath_to_save_audio_project)
                else:
                    root_dirpath_to_save_audio_project = prompt_user_to_select_root_dirpath_to_save_audio_project()

            CliCache().cache().put("root_dirpath_to_save_audio_project", root_dirpath_to_save_audio_project)
            CliCache().save_cache_to_yaml()

        return root_dirpath_to_save_audio_project

    def _content_elements_to_audacity(self, content_elements: List[ContentElement]):
        raise Exception(f"_content_elements_to_audacity function has been deprecated when the "
                        f"inoft_vocal_framework and the inoft_vocal_engine have been splitted.")

        from inoft_vocal_engine.audacity.text_to_audacity import TextToAudacity
        client = TextToAudacity(character_names_to_voices=self.settings.characters_voices)
        root_dirpath_to_save_audio_project = self.prompt_get_root_dirpath_to_save_audio()

        num_content_elements = len(content_elements)
        for i, content_element in enumerate(content_elements):
            if isinstance(content_element.dialogues_lines, list) and len(content_element.dialogues_lines) > 0:
                if content_element.id is None:
                    click.echo(f"Warning ! The content element with the following lines : "
                               f"{content_element.dialogues_lines.__dict__} did not had an id, and has not been saved.")
                else:
                    current_project_dirpath = os.path.join(root_dirpath_to_save_audio_project, content_element.id)
                    audacity_project_filepath = os.path.join(current_project_dirpath, f"{content_element.id}_project.aup")
                    wav_final_render_filepath = os.path.join(current_project_dirpath, "final_render.wav")
                    compressed_final_render_filepath = os.path.join(current_project_dirpath, "compressed_final_render.mp3")

                    # client.audacity.new_project()

                    client.synthesize_dialogues_lines_to_project(dialogues_lines=content_element.dialogues_lines,
                                                                 dirpath_to_save_to=current_project_dirpath)

                    client.audacity.save_project(project_output_filepath=audacity_project_filepath)
                    client.audacity.mix_and_render_all_elements()
                    # todo: find a way to combine all of the audio elements, without mixing them in the same audio clip
                    client.audacity.select_all()
                    client.audacity.export(filepath=wav_final_render_filepath)

                    while not os.path.exists(wav_final_render_filepath):
                        time.sleep(0.1)
                        print(f"Waiting for file {wav_final_render_filepath} to be created by Audacity.")

                    from pydub import AudioSegment
                    final_render_audio_segment = AudioSegment.from_file(file=wav_final_render_filepath, format="wav", frame_rate=16000)
                    final_render_audio_segment.export(out_f=compressed_final_render_filepath, format="mp3", bitrate="48k")

                    print(f"Completed file {i}/{num_content_elements}")

    def upload_audio_contents(self):
        bucket_name = "test-vocal-export-inoft"
        region_name = "eu-west-3"

        root_dirpath_to_save_audio_project = self.prompt_get_root_dirpath_to_save_audio()
        for root, dirs, files in os.walk(root_dirpath_to_save_audio_project):
            for name in files:
                current_filepath = os.path.join(root, name)
                s3_object_key_name = current_filepath.replace(f"{root_dirpath_to_save_audio_project}\\", "")
                self.upload_file_to_s3(filepath=current_filepath, object_key_name=s3_object_key_name,
                                  bucket_name=bucket_name, region_name=region_name)
                print(f"Uploaded {current_filepath} to S3 : {bucket_name}/{s3_object_key_name}")




if __name__ == "__main__":
    BotpressCore().generate_audio_contents_from_texts()
    # BotpressCore().upload_audio_contents()



from flask import Blueprint, render_template, request, jsonify
from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow


control_panel_blueprint = Blueprint("control-panel", __name__, template_folder="templates")
audio_editor_read = "{project_name}:audio-editor:read"


@control_panel_blueprint.route("/<account_username>/control-panel")
def control_panel(account_username: str):
    return auth_workflow(
        account_username=account_username, project_url=None,
        required_permissions=[],
        success_handler=_control_panel, to_template=True
    )

def _control_panel(engine_resources: EngineResources, account_resources: AccountResources, project_resources=None):

    projects = engine_resources.accounts_data_dynamodb_client.get_all_owned_projects_of_account_by_account_username(
        account_username="robinsonlabourdette"
    )
    # todo: fix this fox control panel
    return render_template("control-panel/index.html", account_username="robinsonlabourdette", projects=projects)


@control_panel_blueprint.route("/<account_username>/projects/create", methods=["POST"])
def create_project(account_username: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=None,
        required_permissions=[],
        success_handler=_create_project, to_json=True
    )

def _create_project(engine_resources: EngineResources, account_resources: AccountResources, project_resources=None):
    try:
        from inoft_vocal_engine.models.projects.backend_received_project_data import BackendReceivedProjectData
        project_data = BackendReceivedProjectData(**request.get_json())

        from inoft_vocal_engine.databases.dynamodb.accounts_data_dynamodb_client import ProjectDatabaseItem
        project_database_item = ProjectDatabaseItem(**project_data.dict())

        success = engine_resources.accounts_data_dynamodb_client.add_project_to_account(
            account_id=account_resources.account_id, project_item=project_database_item
        )
        project_complete_url = f"/{account_resources.account_username}/{project_database_item.projectPrimaryUrl}/contents"
        return jsonify(success=success, redirect=project_complete_url)
    except Exception as e:
        print(e)
    return jsonify(success=False)



from typing import List, Dict
from inoft_vocal_engine.databases.dynamodb.projects_diagrams_data_dynamodb_client import ProjectsDiagramsDataDynamoDbClient
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.import_integrations.botpress.engine_models import ProcessedNodeElement, DatabaseTransitionConditionItem


def botpress_diagrams_json_object_to_list_content_elements(list_all_node_elements: list) -> List[ProcessedNodeElement]:
    nodes_names_to_instances: Dict[str, dict] = dict()
    for node_data in list_all_node_elements:
        current_node_name = SafeDict(node_data).get("name").to_str(default=None)
        if current_node_name is not None:
            nodes_names_to_instances[current_node_name] = node_data

    node_elements: List[ProcessedNodeElement] = list()
    for node_data in list_all_node_elements:
        current_node_safedict = SafeDict(node_data)
        current_node_object = ProcessedNodeElement()

        current_node_object.node_id = current_node_safedict.get("id").to_str(default=None)
        current_node_object.node_name = current_node_safedict.get("name").to_str(default=None)
        current_node_object.node_type = current_node_safedict.get("type").to_str(default=None)
        current_node_object.actions = current_node_safedict.get("onEnter").to_list(default=None)
        current_node_object.temp_on_receive = current_node_safedict.get("onReceive").to_list(default=None)

        next_items = current_node_safedict.get("next").to_list(default=None)
        if next_items is not None:
            next_instances: List[DatabaseTransitionConditionItem] = list()
            for next_item in next_items:
                current_next_item_safedict = SafeDict(next_item)
                current_next_item_condition = current_next_item_safedict.get('condition').to_str(default=None)

                current_next_item_target_node_name = current_next_item_safedict.get("node").to_str(default=None)
                if current_next_item_target_node_name in nodes_names_to_instances.keys():
                    current_next_item_target_node_safedict = SafeDict(nodes_names_to_instances[current_next_item_target_node_name])
                    current_next_item_outPortNodeId = current_next_item_target_node_safedict.get("id").to_str(default=None)
                else:
                    current_next_item_outPortNodeId = None

                next_instances.append(DatabaseTransitionConditionItem(condition=current_next_item_condition,
                                                    targetNodeId=current_next_item_outPortNodeId))
            current_node_object.next = next_instances
        node_elements.append(current_node_object)
    return node_elements

def botpress_diagrams_json_file_to_list_content_elements(filepath: str) -> List[ProcessedNodeElement]:
    from inoft_vocal_framework.utils.general import load_json
    return botpress_diagrams_json_object_to_list_content_elements(list_all_node_elements=load_json(filepath))


def put_new_botpress_diagrams_data_from_list(nodes_list: list, projects_diagrams_data_dynamodb_client: ProjectsDiagramsDataDynamoDbClient):
    from inoft_vocal_engine.databases.dynamodb.projects_diagrams_data_dynamodb_client import NodeDatabaseItem

    nodes_list = botpress_diagrams_json_object_to_list_content_elements(list_all_node_elements=nodes_list)
    nodes_names_to_instances = dict()
    for node in nodes_list:
        nodes_names_to_instances[node.nodeName] = node

    for node in nodes_list:
        # todo: make ids dynamic
        db_node_item = NodeDatabaseItem(accountId="79120c34-b345-4718-b9b3-6ffece48ddea",
                                        accountProjectId="5aedb52e-3431-42dc-bf0b-ccd2844d2607",
                                        accountProjectInstanceId="test",
                                        nodeId=node.nodeId, nodeName=node.nodeName, actions=node.actions,
                                        transitions=node.transitions, temp_on_receive=node.temp_on_receive, nodeType=node.nodeType)
        projects_diagrams_data_dynamodb_client.put_new_node(node_item=db_node_item)

def put_new_botpress_diagrams_data_from_filepath(filepath: str, projects_diagrams_data_dynamodb_client: ProjectsDiagramsDataDynamoDbClient):
    nodes_list = botpress_diagrams_json_file_to_list_content_elements(filepath=filepath)
    put_new_botpress_diagrams_data_from_list(nodes_list=nodes_list, projects_diagrams_data_dynamodb_client=projects_diagrams_data_dynamodb_client)



import os
from dataclasses import dataclass
from datetime import datetime
from tempfile import TemporaryFile
from typing import Optional, List

import inoft_vocal_engine
from inoft_vocal_engine.cloud_providers.aws.aws_core import AwsCore
from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.safe_dict import SafeDict
from uuid import uuid4
from inoft_vocal_engine.utils.paths import get_engine_temp_path


@dataclass
class RetrievedS3File:
    Key: str
    LastModified: datetime
    ETag: str
    Size: float
    StorageClass: str
    FileContent: Optional[str] = None


class GeneratedCodeS3Client(AwsCore):
    PROJECTS_DATA_V1_BUCKET_NAME = "inoft-vocal-engine-web-test"

    def __init__(self):
        super().__init__(clients_to_load=[AwsCore.CLIENT_S3])
        self.temp_dirpath = get_engine_temp_path()

    def _make_new_temp_filepath(self) -> os.path:
        # We use a UUID in the filepath, because otherwise, if the name of the temp_app_code file was unique,
        # if the code is executed at the same time by multiple users, the temporary file of one user that is
        # being uploaded and has not yet been deleted, might be overridden by the temporary file of another user.
        temp_filepath = os.path.join(self.temp_dirpath, f"temp_app_code_{str(uuid4())}.py")
        while os.path.exists(temp_filepath):
            # If the filepath already exists (which would be weird since we use an unique uuid in
            # the filepath), we generate a new temp_filepath until we found one that do not yet exists.
            temp_filepath = self._make_new_temp_filepath()
            print(f"Temp path is a duplicate. Creating a new one.\n  --temp_filepath:{temp_filepath}")
        return temp_filepath

    def _make_new_temp_build_dirpath(self, project_resources: ProjectResources) -> os.path:
        temp_build_dirpath = os.path.join(
            self.temp_dirpath, project_resources.project_owner_account_id, project_resources.account_project_id, "builds", str(uuid4())
        )
        while os.path.exists(temp_build_dirpath):
            temp_build_dirpath = self._make_new_temp_build_dirpath(project_resources=project_resources)
            print(f"Build dir path is a duplicate. Creating a new one.\n  --temp_build_dirpath:{temp_build_dirpath}")
        return temp_build_dirpath

    @staticmethod
    def _make_s3_project_folder_key_name(project_resources: ProjectResources) -> str:
        return f"{project_resources.project_owner_account_id}/{project_resources.account_project_id}"

    @staticmethod
    def _make_archive_filename_from_s3_key(project_resources: ProjectResources, s3_key: str) -> str:
        filepath_before_and_with_account_id, filepath_after_account_id = (
            s3_key.split(project_resources.project_owner_account_id, maxsplit=1)
        )
        if len(filepath_after_account_id) == 0:
            filepath_after_account_id = filepath_before_and_with_account_id

        filepath_before_and_with_project_id, filepath_after_project_id = (
            filepath_after_account_id.split(project_resources.account_project_id, maxsplit=1)
        )
        if len(filepath_after_project_id) == 0:
            filepath_after_project_id = filepath_before_and_with_project_id

        filepath_before_and_with_files_folder, filepath_after_files_folder = (
            filepath_after_project_id.split("files", maxsplit=1)
        )
        if len(filepath_after_files_folder) == 0:
            filepath_after_files_folder = filepath_before_and_with_files_folder

        filepath_after_files_folder: str
        filepath_after_files_folder = filepath_after_files_folder.lstrip("/")

        return filepath_after_files_folder

    def upload_app_code(self, project_resources: ProjectResources, app_code: str, key_filename: str):
        # todo: replace the use of real temporary file, by the tempfile module
        temp_dirpath = os.path.join(os.path.dirname(os.path.realpath(inoft_vocal_engine.__file__)), "tmp")
        if not os.path.isdir(temp_dirpath):
            os.makedirs(temp_dirpath)

        temp_filepath = self._make_new_temp_filepath()

        with open(temp_filepath, mode="w+", encoding="utf-8") as temp_file:
            temp_file.write(app_code)

        # todo: make object key name dynamic based on account and project id
        s3_object_key_name = f"{self._make_s3_project_folder_key_name(project_resources=project_resources)}/files/{key_filename}"
        self.upload_file_to_s3(
            filepath=temp_filepath, object_key_name=s3_object_key_name,
            bucket_name=self.PROJECTS_DATA_V1_BUCKET_NAME, region_name="eu-west-3",
            access_should_be_public=True
        )
        # When the temporary file is uploaded to S3, we remove it.
        os.remove(temp_filepath)

    def get_code_of_one_file(self, file_bucket_key: str) -> Optional[str]:
        # todo: optimize this shit code (use safedict, make more precise catches, etc)
        response = self.s3_client.get_object(Bucket=self.PROJECTS_DATA_V1_BUCKET_NAME, Key=file_bucket_key)
        if isinstance(response, dict):
            body = response.get('Body', None)
            if body is not None:
                try:
                    body = body.read().decode('utf-8')
                    return body
                except Exception as e:
                    print(e)
        return None

    def get_infos_of_all_build_files(self, project_resources: ProjectResources) -> Optional[List[RetrievedS3File]]:
        s3_project_folder_key_name = f"{self._make_s3_project_folder_key_name(project_resources=project_resources)}/files"
        response = self.get_all_files_from_s3_folder(
            bucket_name=self.PROJECTS_DATA_V1_BUCKET_NAME,
            folderpath=s3_project_folder_key_name
        )
        if isinstance(response, dict):
            contents: Optional[List[dict]] = response.get('Contents', None)
            if contents is not None:
                output_retrieved_files_objects: List[RetrievedS3File] = list()
                for item in contents:
                    output_retrieved_files_objects.append(RetrievedS3File(**item))
                return output_retrieved_files_objects
        return None

    def get_contents_of_all_build_files(self, project_resources: ProjectResources) -> Optional[List[RetrievedS3File]]:
        files_data = self.get_infos_of_all_build_files(project_resources=project_resources)
        if files_data is not None:
            # At this point, the files_data only contains the infos of each file without its data.
            for file_info in files_data:
                if file_info.Key is not None:
                    file_content = self.get_code_of_one_file(file_bucket_key=file_info.Key)
                    if file_content is not None:
                        file_info.FileContent = file_content

            # After our loop, if we have been able to retrieve the contents,
            # the files_data object also contains the content of each file.
            return files_data
        return None

    def generate_zip_from_build_files(self, project_resources: ProjectResources) -> Optional[str]:
        from tempfile import SpooledTemporaryFile
        import zipfile

        files_data_with_content = self.get_contents_of_all_build_files(project_resources=project_resources)
        if files_data_with_content is not None:
            build_zip_file_key_name = f"{self._make_s3_project_folder_key_name(project_resources=project_resources)}/builds/build1.zip"

            temp_build_dirpath = self._make_new_temp_build_dirpath(project_resources=project_resources)
            temp_build_files_dirpath = os.path.join(temp_build_dirpath, "files")
            os.makedirs(temp_build_files_dirpath)
            # This single makedirs will create the dirs for the temp_build_dirpath and temp_build_files_dirpath variables.
            temp_build_zipfile_filepath = os.path.join(temp_build_dirpath, "build.zip")

            success = False
            try:
                with open(temp_build_zipfile_filepath, mode="w+") as temp_zipfile:
                    with zipfile.ZipFile(temp_build_zipfile_filepath, 'w', zipfile.ZIP_DEFLATED) as archive:
                        for file in files_data_with_content:
                            if file.FileContent is not None:
                                archive_filename = self._make_archive_filename_from_s3_key(project_resources=project_resources, s3_key=file.Key)

                                temp_archive_nested_file_filename = os.path.join(temp_build_files_dirpath, archive_filename)
                                with open(temp_archive_nested_file_filename, "w+") as temp_nested_file:
                                    # For some reasons, the \r in the FileContent do not pose any issues in the string
                                    # representation of PyCharm, but will double each new line when writing the data to a file.
                                    # Removing all the \r fix the issue without causing any loss of integrity to the file data.
                                    temp_nested_file.write(file.FileContent.replace("\r", ""))
                                    # os.chmod(temp_archive_nested_file_filename, 0o777)  # Not necessary to set the file permissions

                                archive.write(filename=temp_archive_nested_file_filename, arcname=archive_filename)

                    # Reset file pointer
                    temp_zipfile.seek(0)

                # os.chmod(temp_build_zipfile_filepath, 0o777)  # Not necessary to set the file permissions
                success = self.upload_file_to_s3(
                    filepath=temp_build_zipfile_filepath, object_key_name=build_zip_file_key_name,
                    bucket_name=self.PROJECTS_DATA_V1_BUCKET_NAME, region_name="eu-west-3",
                    access_should_be_public=True
                )
            except Exception as e:
                print(e)
            finally:
                from shutil import rmtree
                # Delete the folder and all its content
                # (os.remove is only done to remove a single file)
                rmtree(temp_build_dirpath)

            print(f"success = {success}")
            if success is True:
                return build_zip_file_key_name
        return None


if __name__ == "__main__":
    engine_resources = EngineResources()
    account_resources = AccountResources(engine_resources=engine_resources)
    account_resources.account_username = "robinsonlabourdette"
    project_resources_ = ProjectResources(engine_resources=engine_resources, account_resources=account_resources,
                                         project_url="anvers1944", project_owner_account_username="robinsonlabourdette",
                                         project_owner_account_id="b1fe5939-032b-462d-92e0-a942cd445096")

    GeneratedCodeS3Client().generate_zip_from_build_files(project_resources=project_resources_)


"""
                            # with TemporaryFile() as tmp2:
                                # tmp2.write(bytes(file.FileContent))
                                # archive.write(filename=tmp2.name)
                            # todo: fix issue where the files inside the zip are saved in readonly mode, and so
                            #  will not be able to be accessed by Lambda, until we manually modify it in the console."""


import os
from dataclasses import dataclass
from pydantic import BaseModel
from typing import List, Dict
# from typing import TypedDict
from pydantic.class_validators import ROOT_KEY, validator

from inoft_vocal_engine.utils.general import load_json
from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path


# Init of static variables
# todo: upgrade from Python 3.7 to Python 3.8 and re-use the TypedDict
class AlexaCountryObject:  # (TypedDict):
    name: str
    code: str

alexa_countries: List[AlexaCountryObject] = load_json(os.path.join(get_inoft_vocal_engine_root_path(), "web_interface/static/deployment/alexaCountries.json"))
alexa_countries_keys_codes: List[str] = list()
for country in alexa_countries:
    alexa_countries_keys_codes.append(country["code"])


# todo: upgrade from Python 3.7 to Python 3.8 and re-use the TypedDict
class AlexaLocalObject:  # (TypedDict):
    keyName: str
    code: str

alexa_locales: List[AlexaLocalObject] = load_json(os.path.join(get_inoft_vocal_engine_root_path(), "web_interface/static/deployment/alexaLocales.json"))
alexa_locales_key_codes: List[str] = list()
for local in alexa_locales:
    alexa_locales_key_codes.append(local["code"])


alexa_skill_categories: List[str] = load_json(os.path.join(get_inoft_vocal_engine_root_path(), "web_interface/static/deployment/alexaSkillCategories.json"))


# Utils functions
def raise_if_local_key_invalid(localKeyCode: str):
    if not isinstance(localKeyCode, str):
        raise Exception(f"LocalKeyCode was not of string instance."
                        f"\n  --localKeyCode:{localKeyCode}"
                        f"\n  --type(localKeyCode):{type(localKeyCode)}")
    if localKeyCode not in alexa_locales_key_codes:
        raise Exception(f"LocalKeyCode was an invalid key code."
                        f"\n  --localKeyCode:{localKeyCode}"
                        f"\n  --alexa_locales_key_codes:{alexa_locales_key_codes}")


# Requests data models
class ApplicationManifestGetRequestData(BaseModel):
    typeKeyName: str
    data: dict

class ApplicationManifestUpdateRequestData(BaseModel):
    typeKeyName: str
    data: dict


# Custom validation models
class BaseDataModel(BaseModel):
    def __init__(self, database_path: Dict[str, type], **kwargs):
        super().__init__(**kwargs)
        self.__dict__["database_path"] = database_path
        # We use the __dict__ function to set this attribute, so that
        # pydantic will not attempt to validate the database path.

    def export(self):
        # We need a custom export function, because the pydantic dict() will render us a dict of the object, even if
        # the object has only the __root__ key (in which case, we just want the __root__ as the exported value).
        # And the json() function, which will correctly give us only the __root__ variable, will try to serialize
        # any string, and so, it will try to add junk " or ' to the start and end of the string. Here is our solution.
        data = self.dict(exclude={"database_path"})
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        return data

class BaseLocalizedDataModel(BaseDataModel):
    def __init__(self, localKeyCode: str, database_path: Dict[str, type], **kwargs):
        super().__init__(database_path=database_path, **kwargs)
        raise_if_local_key_invalid(localKeyCode=localKeyCode)
        # todo: replace raise Exception by a non-crashing validation error


# Fields getters, updaters and senders data models

# Alexa specific diffusion countries models
class AlexaSkillDiffusionCountriesModel(BaseDataModel):
    def __init__(self, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "diffusionCountries": dict, "selectedCountriesCodes": list}
        super().__init__(database_path=database_path, **kwargs)

class AlexaSkillDiffusionCountriesUpdateModel(AlexaSkillDiffusionCountriesModel):
    __root__: List[str]

    def __init__(self, selectedCountriesCodes: List[str]):
        super().__init__(__root__=selectedCountriesCodes)

    @validator('__root__', each_item=True)
    def check_countries_key_codes(cls, countryKeyCode: str):
        assert countryKeyCode in alexa_countries_keys_codes, (
            f"CountryKeyCode was an invalid key code."
            f"\n  --countryKeyCode:{countryKeyCode}"
            f"\n  --alexa_countries_key_codes:{alexa_countries_keys_codes}")
        return countryKeyCode

class AlexaSkillDiffusionCountriesSenderModel(BaseModel):
    selectedCountriesCodes: List[str]


# Alexa specific test instructions models
class AlexaSkillTestInstructionsModel(BaseDataModel):
    def __init__(self, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "testInstructions": str}
        super().__init__(database_path=database_path, **kwargs)

class AlexaSkillTestInstructionsUpdateModel(AlexaSkillTestInstructionsModel):
    __root__: str

    def __init__(self, testInstructions: str):
        super().__init__(__root__=testInstructions)

class AlexaSkillTestInstructionsSenderModel(BaseModel):
    testInstructions: str


# Alexa specific test instructions models
class AlexaSkillCategoryModel(BaseDataModel):
    def __init__(self, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "skillCategory": str}
        super().__init__(database_path=database_path, **kwargs)

class AlexaSkillCategoryUpdateModel(AlexaSkillCategoryModel):
    __root__: str

    def __init__(self, skillCategory: str):
        super().__init__(__root__=skillCategory)

    @validator('__root__')
    def check_skill_category(cls, skillCategory: str):
        assert skillCategory in alexa_skill_categories, (
            f"SkillCategory was an invalidation category."
            f"\n  --skillCategory:{skillCategory}"
            f"\n  --alexa_skill_categories:{alexa_skill_categories}")
        return skillCategory

class AlexaSkillCategorySenderModel(BaseModel):
    skillCategory: str


# Alexa specific localized skill name models
class AlexaSpecificLocalesSkillNameModel(BaseLocalizedDataModel):
    def __init__(self, localKeyCode: str, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "locales": dict, localKeyCode: dict, "skillName": str}
        super().__init__(localKeyCode=localKeyCode, database_path=database_path, **kwargs)

class AlexaSpecificLocalesSkillNameUpdateModel(AlexaSpecificLocalesSkillNameModel):
    __root__: str

    def __init__(self, localKeyCode: str, skillName: str):
        super().__init__(localKeyCode=localKeyCode, __root__=skillName)

class AlexaSpecificLocalesSkillNameSenderModel(BaseModel):
    skillName: str


# Alexa specific localized skill summary models
class AlexaSpecificLocalesSkillSummaryModel(BaseLocalizedDataModel):
    def __init__(self, localKeyCode: str, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "locales": dict, localKeyCode: dict, "skillSummary": str}
        super().__init__(localKeyCode=localKeyCode, database_path=database_path, **kwargs)

class AlexaSpecificLocalesSkillSummaryUpdateModel(AlexaSpecificLocalesSkillSummaryModel):
    __root__: str

    def __init__(self, localKeyCode: str, skillSummary: str):
        super().__init__(localKeyCode=localKeyCode, __root__=skillSummary)

class AlexaSpecificLocalesSkillSummarySenderModel(BaseModel):
    skillSummary: str


# Alexa specific localized skill description models
class AlexaSpecificLocalesSkillDescriptionModel(BaseLocalizedDataModel):
    def __init__(self, localKeyCode: str, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "locales": dict, localKeyCode: dict, "skillDescription": str}
        super().__init__(localKeyCode=localKeyCode, database_path=database_path, **kwargs)

class AlexaSpecificLocalesSkillDescriptionUpdateModel(AlexaSpecificLocalesSkillDescriptionModel):
    __root__: str

    def __init__(self, localKeyCode: str, skillDescription: str):
        super().__init__(localKeyCode=localKeyCode, __root__=skillDescription)

class AlexaSpecificLocalesSkillDescriptionSenderModel(BaseModel):
    skillDescription: str


# Alexa specific localized skill privacy policy models
class AlexaSpecificLocalesPrivacyPolicyModel(BaseLocalizedDataModel):
    def __init__(self, localKeyCode: str, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "locales": dict, localKeyCode: dict, "skillPrivacyPolicyUrl": str}
        super().__init__(localKeyCode=localKeyCode, database_path=database_path, **kwargs)

class AlexaSpecificLocalesPrivacyPolicyUpdateModel(AlexaSpecificLocalesPrivacyPolicyModel):
    __root__: str

    def __init__(self, localKeyCode: str, skillPrivacyPolicyUrl: str):
        super().__init__(localKeyCode=localKeyCode, __root__=skillPrivacyPolicyUrl)

class AlexaSpecificLocalesPrivacyPolicySenderModel(BaseModel):
    skillPrivacyPolicyUrl: str
    
    
# Alexa specific localized skill terms of use models
class AlexaSpecificLocalesTermsOfUseModel(BaseLocalizedDataModel):
    def __init__(self, localKeyCode: str, **kwargs):
        database_path = {"publishing": dict, "alexa": dict, "locales": dict, localKeyCode: dict, "skillTermsOfUseUrl": str}
        super().__init__(localKeyCode=localKeyCode, database_path=database_path, **kwargs)

class AlexaSpecificLocalesTermsOfUseUpdateModel(AlexaSpecificLocalesTermsOfUseModel):
    __root__: str

    def __init__(self, localKeyCode: str, skillTermsOfUseUrl: str):
        super().__init__(localKeyCode=localKeyCode, __root__=skillTermsOfUseUrl)

class AlexaSpecificLocalesTermsOfUseSenderModel(BaseModel):
    skillTermsOfUseUrl: str


# Switch statements
@dataclass
class SwitchValidatorsWrapper:
    base: any
    update: any
    sender: any

switch_key_type_to_data_models: Dict[str, SwitchValidatorsWrapper] = {
    "AlexaSkillDiffusionCountries": SwitchValidatorsWrapper(
        base=AlexaSkillDiffusionCountriesModel,
        update=AlexaSkillDiffusionCountriesUpdateModel,
        sender=AlexaSkillDiffusionCountriesSenderModel
    ),
    "AlexaSkillTestInstructions": SwitchValidatorsWrapper(
        base=AlexaSkillTestInstructionsModel,
        update=AlexaSkillTestInstructionsUpdateModel,
        sender=AlexaSkillTestInstructionsSenderModel
    ),
    "AlexaSkillCategory": SwitchValidatorsWrapper(
        base=AlexaSkillCategoryModel,
        update=AlexaSkillCategoryUpdateModel,
        sender=AlexaSkillCategorySenderModel
    ),
    "AlexaLocalizedName": SwitchValidatorsWrapper(
        base=AlexaSpecificLocalesSkillNameModel,
        update=AlexaSpecificLocalesSkillNameUpdateModel,
        sender=AlexaSpecificLocalesSkillNameSenderModel
    ),
    "AlexaLocalizedSummary": SwitchValidatorsWrapper(
        base=AlexaSpecificLocalesSkillSummaryModel,
        update=AlexaSpecificLocalesSkillSummaryUpdateModel,
        sender=AlexaSpecificLocalesSkillSummarySenderModel
    ),
    "AlexaLocalizedDescription": SwitchValidatorsWrapper(
        base=AlexaSpecificLocalesSkillDescriptionModel,
        update=AlexaSpecificLocalesSkillDescriptionUpdateModel,
        sender=AlexaSpecificLocalesSkillDescriptionSenderModel
    ),
    "AlexaSkillLocalizedPrivacyPolicy": SwitchValidatorsWrapper(
        base=AlexaSpecificLocalesPrivacyPolicyModel,
        update=AlexaSpecificLocalesPrivacyPolicyUpdateModel,
        sender=AlexaSpecificLocalesPrivacyPolicySenderModel
    ),
    "AlexaSkillLocalizedTermsOfUse": SwitchValidatorsWrapper(
        base=AlexaSpecificLocalesTermsOfUseModel,
        update=AlexaSpecificLocalesTermsOfUseUpdateModel,
        sender=AlexaSpecificLocalesTermsOfUseSenderModel
    ),
}





from flask import Blueprint, render_template, Response, request, jsonify

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow
from inoft_vocal_engine.web_interface.templates.topbar import topbar_ui
from inoft_vocal_engine.web_interface.templates.local_toolbars import ui as local_toolbars_ui

contents_blueprint = Blueprint("contents", __name__, template_folder="templates")

# todo: make it so that when we update the content list, we also reset up the text fields
#  (because currently, they are not re-setup when the html is being replaced)
# todo: make the search capital case insensitive
# todo: fix issue where the characterNames list in the database will only be set when importing from botpress,
#  but not set when updating or adding a new content (even if currently, its only possible to update, not to add)

@contents_blueprint.route("/<account_username>/<project_url>/contents")
def contents(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_contents, to_template=True
    )

def _contents(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.web_interface.static.ui import SIDEBAR_BUTTONS
    return render_template("index.html", account_username=account_resources.account_username,
                           project_url=project_resources.project_url,
                           list_content=_get_latest_updated_contents(project_resources=project_resources)["html"],
                           sidebar_buttons=SIDEBAR_BUTTONS,
                           topbar_buttons=topbar_ui.render(
                               account_username=account_resources.account_username,
                               project_url=project_resources.project_url,
                               active_button_id=topbar_ui.ID_CONTENT
                           ),
                           local_topbar_buttons=local_toolbars_ui.CONTENT_LOCAL_TOOLBAR_BUTTONS)


@contents_blueprint.route("/<account_username>/<project_url>/contents/upload", methods=["POST"])
def upload_post_contents(account_username: str, project_url: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)

    request_json_data = SafeDict(request.get_json())
    content_elements = request_json_data.get("contentElements").to_list(default=None)
    if content_elements is not None:
        from inoft_vocal_engine.import_integrations.botpress.content_elements import put_new_botpress_contents_from_list
        put_new_botpress_contents_from_list(contents_list=content_elements, project_resources=project_resources)

    return Response(status=200)

@contents_blueprint.route("/<account_username>/<project_url>/contents/update/<element_id>", methods=["POST"])
def update_text_content(account_username: str, project_url: str, element_id: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)

    request_json_data = SafeDict(request.get_json())
    print(request_json_data)
    return Response(status=200)

    dialogue_line_index = request_json_data.get("dialogueLineIndex").to_int(default=None)
    element_text = request_json_data.get("text").to_str(default=None)

    project_resources.project_text_contents_dynamodb_client.update_content_element(
        element_id=element_id, dialogue_line_index=dialogue_line_index, element_text=element_text)

    return Response(status=200)

@contents_blueprint.route("/<account_username>/<project_url>/contents/latest-updated", methods=["GET"])
def get_latest_updated_contents(account_username: str, project_url: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)
    return _get_latest_updated_contents(project_resources=project_resources)

def _get_latest_updated_contents(project_resources: ProjectResources):
    content_items = project_resources.project_text_contents_dynamodb_client.get_latest_updated(num_latest_items=30).items
    # todo: make latest contents customizable in the backend and in the frontend
    return {"html": render_template("list_content_elements/list.html", elements=content_items)}

@contents_blueprint.route("/<account_username>/<project_url>/contents/characterName", methods=["GET"])
def get_contents_by_character_name(account_username: str, project_url: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)

    character_name = request.args.get('inputValue', default=None, type=str)
    if character_name is not None:
        database_response = project_resources.project_text_contents_dynamodb_client.get_by_character_name(character_name=character_name)
        if database_response.items is not None and len(database_response.items) > 0:
            return jsonify({"html": render_template("list_content_elements/list.html", elements=database_response.items)})

    return jsonify({"html": "<p>Aucun résultats trouvés</p>"})

@contents_blueprint.route("/<account_username>/<project_url>/contents/text", methods=["GET"])
def get_contents_by_text_search(account_username: str, project_url: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)

    text_to_search = request.args.get('inputValue', default=None, type=str)
    if text_to_search is not None:
        database_response = project_resources.project_text_contents_dynamodb_client.get_by_text_search(
            text_to_search=text_to_search, section_instance_id=12
        )
        # todo: make non-static
        if database_response.items is not None and len(database_response.items) > 0:
            return jsonify({"html": render_template("list_content_elements/list.html", elements=database_response.items)})

    # todo: replace standard text by a template to inform that no results have been found
    return jsonify({"html": "<p>Aucun résultats trouvés</p>"})




from inoft_vocal_engine.permissions.permission_model import PermissionModel

project_diagrams_nodes_read_all = PermissionModel(
    expression="{requester_account_id}:{account_project_id}:diagrams:nodes:read:all",
    display_name="project_diagrams_nodes_read_all", description="Fait des trucs cool"
)
project_diagrams_nodes_update_all = PermissionModel(
    expression="{requester_account_id}:{account_project_id}:diagrams:nodes:update:all",
    display_name="project_diagrams_nodes_update_all", description="lel"
)
project_diagrams_nodes_add_all = PermissionModel(
    expression="{requester_account_id}:{account_project_id}:diagrams:nodes:add:all",
    display_name="project_diagrams_nodes_add_all", description=""
)
project_diagrams_nodes_remove_all = PermissionModel(
    expression="{requester_account_id}:{account_project_id}:diagrams:nodes:remove:all",
    display_name="project_diagrams_nodes_remove_all", description=""
)



from typing import Optional

from flask import Blueprint, render_template, request, jsonify
from pydantic import ValidationError

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.smapi_clients.smapi_base_client import SmapiBaseClient
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow

builds_blueprint = Blueprint("builds", __name__, template_folder="templates")

@builds_blueprint.route("/<account_username>/<project_url>/builds")
def builds(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_builds, to_template=True,
    )

def _builds(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.cloud_providers.aws.deploy import DeployHandler
    build_data = DeployHandler().deploy(engine_resources=engine_resources, project_resources=project_resources, lambda_handler="diagrams_0.py")
    success = engine_resources.accounts_data_dynamodb_client.set_update_infos_of_one_project_build(
        project_resources=project_resources, build=build_data
    )

    return render_template(
        "deployment/deployment.html",
        **project_resources.required_template_kwargs()
    )




# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2019 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

"""Some documentation to refer to:

- Our main web socket (mWS) sends opcode 4 with a guild ID and channel ID.
- The mWS receives VOICE_STATE_UPDATE and VOICE_SERVER_UPDATE.
- We pull the session_id from VOICE_STATE_UPDATE.
- We pull the token, endpoint and server_id from VOICE_SERVER_UPDATE.
- Then we initiate the voice web socket (vWS) pointing to the endpoint.
- We send opcode 0 with the user_id, server_id, session_id and token using the vWS.
- The vWS sends back opcode 2 with an ssrc, port, modes(array) and hearbeat_interval.
- We send a UDP discovery packet to endpoint:port and receive our IP and our port in LE.
- Then we send our IP and port via vWS with opcode 1.
- When that's all done, we receive opcode 4 from the vWS.
- Finally we can transmit data to endpoint:port.
"""

import asyncio
import socket
import logging
import struct
import threading

from . import opus
from .backoff import ExponentialBackoff
from .gateway import *
from .errors import ClientException, ConnectionClosed
from .player import AudioPlayer, AudioSource
from .reader import AudioReader, AudioSink
from .utils import Bidict
from .speakingstate import SpeakingState

try:
    import nacl.secret
    has_nacl = True
except ImportError:
    has_nacl = False


log = logging.getLogger(__name__)

class VoiceClient:
    """Represents a Discord voice connection.

    You do not create these, you typically get them from
    e.g. :meth:`VoiceChannel.connect`.

    Warning
    --------
    In order to use PCM based AudioSources, you must have the opus library
    installed on your system and loaded through :func:`opus.load_opus`.
    Otherwise, your AudioSources must be opus encoded (e.g. using :class:`FFmpegOpusAudio`)
    or the library will not be able to transmit audio.

    Attributes
    -----------
    session_id: :class:`str`
        The voice connection session ID.
    token: :class:`str`
        The voice connection token.
    endpoint: :class:`str`
        The endpoint we are connecting to.
    channel: :class:`abc.Connectable`
        The voice channel connected to.
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop that the voice client is running on.
    """
    def __init__(self, state, timeout, channel):
        if not has_nacl:
            raise RuntimeError("PyNaCl library needed in order to use voice")

        self.channel = channel
        self.main_ws = None
        self.timeout = timeout
        self.ws = None
        self.socket = None
        self.loop = state.loop
        self._state = state
        # this will be used in the AudioPlayer thread
        self._connected = threading.Event()

        self._handshaking = False
        self._handshake_check = asyncio.Lock()
        self._handshake_complete = asyncio.Event()
        # this will be used in the AudioReader thread
        self._connecting = threading.Condition()

        self._mode = None
        self._connections = 0
        self.sequence = 0
        self.timestamp = 0
        self._nonce = 0
        self._runner = None
        self._player = None
        self._reader = None
        self.encoder = None
        self._ssrcs = Bidict()

    warn_nacl = not has_nacl
    supported_modes = (
        'xsalsa20_poly1305_lite',
        'xsalsa20_poly1305_suffix',
        'xsalsa20_poly1305',
    )

    @property
    def guild(self):
        """Optional[:class:`Guild`]: The guild we're connected to, if applicable."""
        return getattr(self.channel, 'guild', None)

    @property
    def user(self):
        """:class:`ClientUser`: The user connected to voice (i.e. ourselves)."""
        return self._state.user

    def checked_add(self, attr, value, limit):
        val = getattr(self, attr)
        if val + value > limit:
            setattr(self, attr, 0)
        else:
            setattr(self, attr, val + value)

    # connection related

    async def start_handshake(self):
        log.info('Starting voice handshake...')

        guild_id, channel_id = self.channel._get_voice_state_pair()
        state = self._state
        self.main_ws = ws = state._get_websocket(guild_id)
        self._connections += 1

        # request joining
        await ws.voice_state(guild_id, channel_id)

        try:
            await asyncio.wait_for(self._handshake_complete.wait(), timeout=self.timeout)
        except asyncio.TimeoutError:
            await self.terminate_handshake(remove=True)
            raise

        log.info('Voice handshake complete. Endpoint found %s (IP: %s)', self.endpoint, self.endpoint_ip)

    async def terminate_handshake(self, *, remove=False):
        guild_id, channel_id = self.channel._get_voice_state_pair()
        self._handshake_complete.clear()
        await self.main_ws.voice_state(guild_id, None, self_mute=True)

        log.info('The voice handshake is being terminated for Channel ID %s (Guild ID %s)', channel_id, guild_id)
        if remove:
            log.info('The voice client has been removed for Channel ID %s (Guild ID %s)', channel_id, guild_id)
            key_id, _ = self.channel._get_voice_client_key()
            self._state._remove_voice_client(key_id)

    async def _create_socket(self, server_id, data):
        async with self._handshake_check:
            if self._handshaking:
                log.info("Ignoring voice server update while handshake is in progress")
                return
            self._handshaking = True

        self._connected.clear()
        self.session_id = self.main_ws.session_id
        self.server_id = server_id
        self.token = data.get('token')
        endpoint = data.get('endpoint')

        if endpoint is None or self.token is None:
            log.warning('Awaiting endpoint... This requires waiting. ' \
                        'If timeout occurred considering raising the timeout and reconnecting.')
            return

        self.endpoint = endpoint.replace(':80', '')
        self.endpoint_ip = socket.gethostbyname(self.endpoint)

        if self.socket:
            try:
                self.socket.close()
            except Exception:
                pass

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)

        if self._handshake_complete.is_set():
            # terminate the websocket and handle the reconnect loop if necessary.
            self._handshake_complete.clear()
            await self.ws.close(4000)
            return

        self._handshake_complete.set()

    async def connect(self, *, reconnect=True, _tries=0, do_handshake=True):
        log.info('Connecting to voice...')
        try:
            del self.secret_key
        except AttributeError:
            pass

        if do_handshake:
            await self.start_handshake()

        try:
            self.ws = await DiscordVoiceWebSocket.from_client(self)
            self._handshaking = False
            self._connected.clear()
            while not hasattr(self, 'secret_key'):
                await self.ws.poll_event()
            self._connected.set()
        except (ConnectionClosed, asyncio.TimeoutError):
            if reconnect and _tries < 5:
                log.exception('Failed to connect to voice... Retrying...')
                await asyncio.sleep(1 + _tries * 2.0)
                await self.terminate_handshake()
                await self.connect(reconnect=reconnect, _tries=_tries + 1)
            else:
                raise

        if self._runner is None:
            self._runner = self.loop.create_task(self.poll_voice_ws(reconnect))

    async def poll_voice_ws(self, reconnect):
        backoff = ExponentialBackoff()
        while True:
            try:
                await self.ws.poll_event()
            except (ConnectionClosed, asyncio.TimeoutError) as exc:
                if isinstance(exc, ConnectionClosed):
                    # The following close codes are undocumented so I will document them here.
                    # 1000 - normal closure (obviously)
                    # 4014 - voice channel has been deleted.
                    # 4015 - voice server has crashed
                    if exc.code in (1000, 4014, 4015):
                        log.info('Disconnecting from voice normally, close code %d.', exc.code)
                        await self.disconnect()
                        break

                if not reconnect:
                    await self.disconnect()
                    raise

                retry = backoff.delay()
                log.exception('Disconnected from voice... Reconnecting in %.2fs.', retry)
                self._connected.clear()
                await asyncio.sleep(retry)
                await self.terminate_handshake()
                try:
                    await self.connect(reconnect=True)
                except asyncio.TimeoutError:
                    # at this point we've retried 5 times... let's continue the loop.
                    log.warning('Could not connect to voice... Retrying...')
                    continue

    async def disconnect(self, *, force=False):
        """|coro|

        Disconnects this voice client from voice.
        """
        if not force and not self.is_connected():
            return

        self.stop()
        self._connected.clear()

        try:
            if self.ws:
                await self.ws.close()

            await self.terminate_handshake(remove=True)
        finally:
            if self.socket:
                self.socket.close()

    async def move_to(self, channel):
        """|coro|

        Moves you to a different voice channel.

        Parameters
        -----------
        channel: :class:`abc.Snowflake`
            The channel to move to. Must be a voice channel.
        """
        guild_id, _ = self.channel._get_voice_state_pair()
        await self.main_ws.voice_state(guild_id, channel.id)

    def is_connected(self):
        """Indicates if the voice client is connected to voice."""
        return self._connected.is_set()

    async def speak(self, state):
        """|coro|

        Sets the bot's speaking state.
        [maybe note about how the bot does this while playing audio so
        you probably dont need to call this]

        Parameters
        -----------
        state: :class:`SpeakingState`
            The state to set the bot to.
        """

        if not isinstance(state, SpeakingState):
            raise TypeError("state must be a SpeakingState not %s" % state.__class__.__name__)

        await self.ws.speak(state)

    # audio related

    def _add_ssrc(self, user_id, ssrc):
        """Adds a user_id<->ssrc mapping.

        Technically these can be added in either order, but using user_id as the key
        is preferable since, should we be updating a mapping instead of adding one, we
        want to update the ssrc for the user_id, not the other way around.

        I think?  Did I write it to work like that?
        """
        self._ssrcs[user_id] = ssrc

    def _remove_ssrc(self, *, ssrc=None, user_id=None):
        """Removes a user_id<->ssrc mapping.  Either one can be used as the key."""

        thing = ssrc or user_id
        if not thing:
            raise TypeError("must provide at least one argument")

        other_thing = self._ssrcs.pop(thing, None)
        if self._reader:
            self._reader._ssrc_removed(ssrc or other_thing)

    def _get_ssrc_mapping(self, *, ssrc=None, user_id=None):
        """Returns a (ssrc, user_id) tuple from the given input.  At least one argument is required."""

        thing = ssrc or user_id
        if not thing:
            raise TypeError("must provide at least one argument")

        other_thing = self._ssrcs.get(thing)
        return ssrc or other_thing, user_id or other_thing

    def _encrypt_voice_packet(self, data):
        header = bytearray(12)

        # Formulate rtp header
        header[0] = 0x80
        header[1] = 0x78
        struct.pack_into('>H', header, 2, self.sequence)
        struct.pack_into('>I', header, 4, self.timestamp)
        struct.pack_into('>I', header, 8, self.ssrc)

        encrypt_packet = getattr(self, '_encrypt_' + self._mode)
        return encrypt_packet(header, data)

    def _encrypt_xsalsa20_poly1305(self, header, data):
        box = nacl.secret.SecretBox(bytes(self.secret_key))
        nonce = bytearray(24)
        nonce[:12] = header

        return header + box.encrypt(bytes(data), bytes(nonce)).ciphertext

    def _encrypt_xsalsa20_poly1305_suffix(self, header, data):
        box = nacl.secret.SecretBox(bytes(self.secret_key))
        nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)

        return header + box.encrypt(bytes(data), nonce).ciphertext + nonce

    def _encrypt_xsalsa20_poly1305_lite(self, header, data):
        box = nacl.secret.SecretBox(bytes(self.secret_key))
        nonce = bytearray(24)
        self.checked_add('_nonce', 1, 4294967295)
        nonce[:4] = self._nonce.to_bytes(4, 'big')

        return header + box.encrypt(bytes(data), bytes(nonce)).ciphertext + nonce[:4]

    def send_audio_packet(self, data, *, encode=True):
        """Sends an audio packet composed of the data.

        You must be connected to play audio.

        Parameters
        ----------
        data: :class:`bytes`
            The :term:`py:bytes-like object` denoting PCM or Opus voice data.
        encode: :class:`bool`
            Indicates if ``data`` should be encoded into Opus.

        Raises
        -------
        ClientException
            You are not connected.
        opus.OpusError
            Encoding the data failed.
        """

        self.checked_add('sequence', 1, 65535)
        if encode:
            if self.encoder is None:
                self.encoder = opus.Encoder()

            encoded_data = self.encoder.encode(data, opus.Encoder.SAMPLES_PER_FRAME)
        else:
            encoded_data = data
        packet = self._encrypt_voice_packet(encoded_data)
        try:
            self.socket.sendto(packet, (self.endpoint_ip, self.voice_port))
        except BlockingIOError:
            log.warning('A packet has been dropped (seq: %s, timestamp: %s)', self.sequence, self.timestamp)

        self.checked_add('timestamp', opus.Encoder.SAMPLES_PER_FRAME, 4294967295)

    # send api related

    def play(self, source, *, after=None):
        """Plays an :class:`AudioSource`.

        The finalizer, ``after`` is called after the source has been exhausted
        or an error occurred.

        If an error happens while the audio player is running, the exception is
        caught and the audio player is then stopped.  If no after callback is
        passed, any caught exception will be displayed as if it were raised.

        Parameters
        -----------
        source: :class:`AudioSource`
            The audio source we're reading from.
        after: Callable[[:class:`Exception`], Any]
            The finalizer that is called after the stream is exhausted.
            This function must have a single parameter, ``error``, that
            denotes an optional exception that was raised during playing.

        Raises
        -------
        ClientException
            Already playing audio or not connected.
        TypeError
            Source is not a :class:`AudioSource` or after is not a callable.
        OpusNotLoaded
            Source is not opus encoded and opus is not loaded.
        """

        if not self.is_connected():
            raise ClientException('Not connected to voice.')

        if self.is_playing():
            raise ClientException('Already playing audio.')

        if not isinstance(source, AudioSource):
            raise TypeError('source must be an AudioSource not {0.__class__.__name__}'.format(source))

        if not self.encoder and not source.is_opus():
            self.encoder = opus.Encoder()

        self._player = AudioPlayer(source, self, after=after)
        self._player.start()

    def is_playing(self):
        """Indicates if we're currently playing audio."""
        return self._player is not None and self._player.is_playing()

    def is_paused(self):
        """Indicates if we're playing audio, but if we're paused."""
        return self._player is not None and self._player.is_paused()

    def stop_playing(self):
        """Stops playing audio."""
        if self._player:
            self._player.stop()
            self._player = None

    def pause(self):
        """Pauses the audio playing."""
        if self._player:
            self._player.pause()

    def resume(self):
        """Resumes the audio playing."""
        if self._player:
            self._player.resume()

    @property
    def source(self):
        """Optional[:class:`AudioSource`]: The audio source being played, if playing.

        This property can also be used to change the audio source currently being played.
        """
        return self._player.source if self._player else None

    @source.setter
    def source(self, value):
        if not isinstance(value, AudioSource):
            raise TypeError('expected AudioSource not {0.__class__.__name__}.'.format(value))

        if self._player is None:
            raise ValueError('Not playing anything.')

        self._player._set_source(value)

    # receive api related

    def listen(self, sink):
        """Receives audio into a :class:`AudioSink`. TODO: wording

        TODO: the rest of it
        """

        if not self.is_connected():
            raise ClientException('Not connected to voice.')

        if not isinstance(sink, AudioSink):
            raise TypeError('sink must be an AudioSink not {0.__class__.__name__}'.format(sink))

        if self.is_listening():
            raise ClientException('Already receiving audio.')

        self._reader = AudioReader(sink, self)
        self._reader.start()

    def is_listening(self):
        """Indicates if we're currently receiving audio."""
        return self._reader is not None and self._reader.is_listening()

    def stop_listening(self):
        """Stops receiving audio."""
        if self._reader:
            self._reader.stop()
            self._reader = None

    @property
    def sink(self):
        return self._reader.sink if self._reader else None

    @sink.setter
    def sink(self, value):
        if not isinstance(value, AudioSink):
            raise TypeError('expected AudioSink not {0.__class__.__name__}.'.format(value))

        if self._reader is None:
            raise ValueError('Not receiving anything.')

        self._reader._set_sink(sink)

    def stop(self):
        """Stops playing and receiving audio."""
        self.stop_playing()
        self.stop_listening()



from .skill_builder.inoft_skill_builder import InoftSkill, InoftCondition, InoftHandler, \
    InoftRequestHandler, InoftStateHandler, InoftDefaultFallback, InoftHandlersGroup, canProcessIntentNames

from .platforms_handlers.simulator.simulator_core import Simulator
from .speechs.ssml_builder_core import Speech, SpeechsList

from .platforms_handlers.discord.handler_input import start_discord_listening

from inoft_vocal_framework.skill_settings.skill_settings import Settings

# todo: if the framework do not find the AWS credentials, ask with a dialog
#  if the developer want to run the app with the database disabled


"""""
function build(intent) {
    console.log(intent);

    let endpoints_json = {
        "action-endpoints": {
        }
    }
    endpoints_json["action-endpoints"].

    /*endpoints {
  action-endpoints {
    action-endpoint (launch) {
      accepted-inputs ($vivContext)
      // local-endpoint (launch.js)
      remote-endpoint ("{remote.url}") {
        method (POST)
      }
    }
    action-endpoint (ANumber) {
      accepted-inputs ($vivContext, IntNumber)
      local-endpoint (ANumber.js)
    }
    action-endpoint (GetContent) {
      accepted-inputs (searchTerm, $vivContext)
      local-endpoint (GetContent.js)
    }
  }
}*/
}"""""
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type, raise_if_value_not_in_list

from inoft_vocal_framework.safe_dict import SafeDict

class ActionEndpoints:
    json_key = "action-endpoints"

    class NestedActionEndpoints:
        json_key = "action-endpoints"

        class RemoteEndpoint:
            json_key = "remote-endpoint"
            METHOD_GET = "GET"
            METHOD_PUT = "PUT"
            METHOD_POST = "POST"
            METHOD_DELETE = "DELETE"
            METHOD_UPDATE = "UPDATE"
            METHOD_OPTIONS = "OPTIONS"
            AVAILABLE_METHODS = [METHOD_GET, METHOD_PUT, METHOD_POST, METHOD_DELETE, METHOD_UPDATE, METHOD_OPTIONS]
            #todo: cannot use nested objects processing, need to use templating

            def __init__(self):
                self.method = None

        def __init__(self):
            self.accepted_inputs = None
            self.local_endpoint = None
            self._remote_endpoint = None

        @property
        def remote_endpoint(self):
            return self._remote_endpoint

        @remote_endpoint.setter
        def remote_endpoint(self, endpoint_url: str, endpoint_method: str = RemoteEndpoint.METHOD_POST) -> None:
            raise_if_variable_not_expected_type(value=endpoint_url, expected_type=endpoint_url, variable_name="endpoint_url")
            raise_if_value_not_in_list(value=endpoint_method, list_object=self.RemoteEndpoint.AVAILABLE_METHODS, variable_name="endpoint_method")
            self._remote_endpoint = self.RemoteEndpoint

        def return_transformations(self):
            vars(self)["accepted-inputs"] = self.accepted_inputs
            del self.accepted_inputs
            vars(self)["local-endpoint"] = self.local_endpoint
            del self.local_endpoint

    def __init__(self):
        self.action_endpoints = self.NestedActionEndpoints()


from inoft_vocal_framework.platforms_handlers.nested_object_to_dict import NestedObjectToDict
endpoints = ActionEndpoints()
dict_object = NestedObjectToDict.get_dict_from_nested_object(object_to_process=ActionEndpoints(),
                                                             key_names_identifier_objects_to_go_into=["json_key"])
print(dict_object)

def build_endpoints():
    endpoints = SafeDict()
    endpoints.get_set("action-endpoints", dict()).get_set("action-endpoints")

# from inoft_vocal_framework.utils.general import load_json
# current_template_infos_dict = load_json(filepath="F:\Inoft\skill_histoire_decryptage\inoft_vocal_framework\bixby_core\dummy_test.json")




from inoft_vocal_framework.bixby_core.templates.templates_access import TemplatesAccess
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.utils.general import load_json

model_dict = SafeDict(load_json("F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_framework/bixby_core/test_model.json"))
intents = model_dict.get("intents").to_dict()

class Core:
    def render(self):
        out = TemplatesAccess().endpoints_template.render(intents=intents)
        print(out)

    @staticmethod
    def write_to_file(text: str, filepath: str):
        with open(filepath, "w+") as file:
            file.write(text)

Core().render()



class StaticData:
    _ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_ID = "amzn1.application-oa2-client.581f5baa50e14800917401bf587f0800"
    _ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_SECRET = "dc1c2b42fc213d19bebc8ef9e78e9ce48ad51620893b0fc2a253ee6a34bac442"
        
    @property
    def ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_ID(self) -> str:
        return StaticData._ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_ID
    
    @property
    def ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_SECRET(self) -> str:
        return self._ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_SECRET



import decimal
from decimal import Decimal, Context, Clamped, Overflow, Inexact, Rounded, Underflow, ROUND_UP
from typing import Any, List


class Utils:
    TYPE_STRING = 'S'
    TYPE_NUMBER = 'N'
    TYPE_BINARY = 'B'
    TYPE_STRING_SET = 'SS'
    TYPE_NUMBER_SET = 'NS'
    TYPE_BINARY_SET = 'BS'
    TYPE_NULL = 'NULL'
    TYPE_BOOLEAN = 'BOOL'
    TYPE_MAP = 'M'
    TYPE_LIST = 'L'
    ALL_TYPES_WHERE_VALUE_DO_NOT_NEED_MODIFICATIONS = [TYPE_STRING, TYPE_BOOLEAN]

    DECIMAL_DYNAMODB_CONTEXT = Context(prec=38, rounding=decimal.ROUND_HALF_EVEN, Emin=-128, Emax=126,
                                       capitals=1, clamp=0, flags=[], traps=[])
    # 38 is the maximum numbers of Decimals numbers that DynamoDB can support.

    from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
    # The serializer and deserializer can be static, even in a lambda,
    # since their classes will never change according to the user.
    _serializer = None
    _deserializer = None

    @classmethod
    def serializer(cls) -> TypeSerializer:
        if cls._serializer is None:
            cls._serializer = cls.TypeSerializer()
        return cls._serializer
    
    @classmethod
    def deserializer(cls) -> TypeDeserializer:
        if cls._deserializer is None:
            cls._deserializer = cls.TypeDeserializer()
        return cls._deserializer

    # todo: deprecate this function that was a complication because i was using the decimal module
    @staticmethod
    def float_serializer(python_object):
        if isinstance(python_object, float):
            # Okay, this is a weird one. The decimal module will crash if it is passed a float
            # with less than two decimal numbers, like float(0.5) which has only one. The decimal
            # module is used when using the Utils.TypeSerializer accessible via the serializer property.
            # Yet, the module has no issues with number with more than 20 decimals, and we want to keep
            # this infos, so we will not round the object. Instead, with the "%.f" % python_object we will
            # get the number of decimal numbers on the float (returned in the form of a string, for real,
            # and if this number is under two, we will turn the float into a float with two decimal numbers by
            # using "%.2f" % python_object, which will then be able to be passed to the Decimal modules without errors.
            # PS : Obviously, instead of doing this, a number like 0.5 cannot be converted to an int ;)
            if ("%.f" % python_object) < "2":
                python_object = "%.2f" % python_object

            return Decimal(python_object)
        elif isinstance(python_object, list):
            for i, item in enumerate(python_object):
                python_object[i] = Utils.float_serializer(python_object=item)
            return python_object
        elif isinstance(python_object, dict):
            for key, value in python_object.items():
                python_object[key] = Utils.float_serializer(python_object=value)
            return python_object
        else:
            return python_object

    def decimal_deserializer(self, python_object):
        from decimal import Decimal
        if isinstance(python_object, Decimal):
            if ("%.f" % python_object) > "0":
                python_object = float(python_object)
            else:
                python_object = int(python_object)
            return python_object
        elif isinstance(python_object, list):
            for i, item in enumerate(python_object):
                python_object[i] = Utils.decimal_deserializer(python_object=item)
            return python_object
        elif isinstance(python_object, dict):
            for key, value in python_object.items():
                python_object[key] = Utils.decimal_deserializer(python_object=value)
            return python_object
        else:
            return python_object

    @staticmethod
    def python_to_dynamodb(python_object: Any):
        """
            None                                    {'NULL': True}
            True/False                              {'BOOL': True/False}
            int/Decimal                             {'N': str(value)}
            string                                  {'S': string}
            Binary/bytearray/bytes (py3 only)       {'B': bytes}
            set([int/Decimal])                      {'NS': [str(value)]}
            set([string])                           {'SS': [string])
            set([Binary/bytearray/bytes])           {'BS': [bytes]}
            list                                    {'L': list}
            dict                                    {'M': dict}
        :param python_object:
        :return:
        """

        object_type = type(python_object)
        if object_type in [int, float]:
            return {Utils.TYPE_NUMBER: Utils._python_to_decimal(python_number=python_object)}
        elif object_type == list:
            for i, item in enumerate(python_object):
                python_object[i] = Utils.python_to_dynamodb(python_object=item)
            return {Utils.TYPE_LIST: python_object}
        elif object_type == dict:
            for key, item in python_object.items():
                python_object[key] = Utils.python_to_dynamodb(python_object=item)
            return {Utils.TYPE_MAP: python_object}
        elif object_type == bool:
            return {Utils.TYPE_BOOLEAN: python_object}
        elif object_type == str:
            return {Utils.TYPE_STRING: python_object}
        elif object_type == type(None):
            return {Utils.TYPE_NULL: True}
        elif object_type == bytes:
            return {Utils.TYPE_BINARY: python_object}
        else:
            return python_object

    @staticmethod
    def dynamodb_to_python(dynamodb_object: Any):
        if isinstance(dynamodb_object, Decimal):
            return Utils._decimal_to_python(decimal_number=dynamodb_object)
        elif isinstance(dynamodb_object, list):
            for i, item in enumerate(dynamodb_object):
                dynamodb_object[i] = Utils.dynamodb_to_python(dynamodb_object=item)
            return dynamodb_object
        elif isinstance(dynamodb_object, dict):
            if len(dynamodb_object) == 1:
                # If the length of the Dict is only one, it might be a DynamoDB object, with its key
                # as its variable type. For example : {'N': '1'}  And yes, this also apply to lists and maps.

                first_key = list(dynamodb_object.keys())[0]
                first_item = dynamodb_object[first_key]

                # First do we thing, is check if the value is a Decimal value (which happens often).
                if isinstance(first_item, Decimal):
                    return Utils._decimal_to_python(decimal_number=first_item)

                # Then the order of the elif statement is based on what we personally use the most at Inoft.
                # For example, we always use the numbers, string, list and dictionaries (maps) but always
                # never binary data or sets. In the default library, the list and dict were at the end of the loop.
                elif first_key == Utils.TYPE_NUMBER:
                    return Utils._dynamodb_number_to_python(number_string=first_item)
                elif first_key == Utils.TYPE_STRING:
                    return first_item
                elif first_key in Utils.ALL_TYPES_WHERE_VALUE_DO_NOT_NEED_MODIFICATIONS:
                    return first_item
                elif first_key == Utils.TYPE_MAP:
                    return dict([(key, Utils.dynamodb_to_python(element)) for key, element in first_item.items()])
                elif first_key == Utils.TYPE_LIST:
                    return [Utils.dynamodb_to_python(element) for element in first_item]
                elif first_key == Utils.TYPE_NULL:
                    return None
                elif first_key == Utils.TYPE_BINARY:
                    return Utils._dynamodb_binary_to_python(binary_data=first_item)
                elif first_key == Utils.TYPE_NUMBER_SET:
                    return set(map(Utils.DECIMAL_DYNAMODB_CONTEXT.create_decimal, first_item))
                elif first_key == Utils.TYPE_STRING_SET:
                    return set(first_item)
                elif first_key == Utils.TYPE_BINARY_SET:
                    from boto3.dynamodb.types import Binary
                    return set(map(Utils._dynamodb_binary_to_python, first_item))

            # If the dict was a classic dict, with its first key not in the keys used by DynamoDB
            keys_to_pop: List[str] = list()
            for key, item in dynamodb_object.items():
                item_value = Utils.dynamodb_to_python(dynamodb_object=item)
                if item_value is not None:
                    dynamodb_object[key] = item_value
                else:
                    keys_to_pop.append(key)

            for key in keys_to_pop:
                dynamodb_object.pop(key)

            return dynamodb_object
        else:
            return dynamodb_object

    @staticmethod
    def _decimal_to_python(decimal_number: Decimal):
        # todo: improve the way the decimal works, because sometimes
        #  a value with no decimal will be converted to a float...
        if ("%.f" % decimal_number) > "0":
            return int(decimal_number)
        else:
            return int(decimal_number)

    @staticmethod
    def _python_to_decimal(python_number: int or float) -> Decimal:
        return Utils.DECIMAL_DYNAMODB_CONTEXT.create_decimal(python_number)

    @staticmethod
    def _dynamodb_number_to_python(number_string: str):
        float_number = float(number_string)
        if float_number.is_integer():
            return int(float_number)
        else:
            return float_number

    @staticmethod
    def _dynamodb_binary_to_python(binary_data):
        from boto3.dynamodb.types import Binary
        return Binary(value=binary_data)

    @staticmethod
    def python_type_to_dynamodb_type(python_type):
        """
        {'NULL': True}                          None
        {'BOOL': True/False}                    True/False
        {'N': str(value)}                       Decimal(str(value))
        {'S': string}                           string
        {'B': bytes}                            Binary(bytes)
        {'NS': [str(value)]}                    set([Decimal(str(value))])
        {'SS': [string]}                        set([string])
        {'BS': [bytes]}                         set([bytes])
        {'L': list}                             list
        {'M': dict}                             dict
        """
        if python_type == type(None):
            return "NULL"
        elif python_type == bool:
            return "BOOL"
        elif python_type in [Decimal, int, float]:
            return "N"
        elif python_type == str:
            return "S"
        elif python_type == bytes:
            return "B"
        elif python_type == list:
            return "L"
        elif python_type == dict:
            return "M"



from pydantic import validate_arguments
from inoft_vocal_engine.databases.dynamodb.accounts_data_dynamodb_client import UsersDynamoDBClient


# todo: make engine resources access static, instead of creating a new class on each lambda invocation ?
class EngineResources:
    _accounts_data_dynamodb_client = None
    _users_dynamodb_table_name = "inoft-vocal-engine_accounts-data"

    @validate_arguments
    def __init__(self):
        pass

    @property
    def accounts_data_dynamodb_client(self) -> UsersDynamoDBClient:
        if self._accounts_data_dynamodb_client is None:
            self._accounts_data_dynamodb_client = UsersDynamoDBClient(
                table_name=self._users_dynamodb_table_name, region_name="eu-west-2"
            )
        return self._accounts_data_dynamodb_client




import json

from flask import Blueprint, render_template, request, Response, jsonify
from pydantic import ValidationError
from typing import List

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow

intents_blueprint = Blueprint("intents", __name__, template_folder="templates")


@intents_blueprint.route("/<account_username>/<project_url>/intents", methods=["GET"])
def intents(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, required_permissions=[],
        success_handler=_intents, to_template=True,
        project_url=project_url
    )

def _intents(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    intents_instances_list = project_resources.vocal_apps_model_schemas_dynamodb_client.get_all_intents_instances_by_account_project_id(
        account_project_id=project_resources.account_project_id
    )
    return render_template("intents/intents_index.html", intents=intents_instances_list,
                           **project_resources.required_template_kwargs())

@intents_blueprint.route("/<account_username>/<project_url>/intents/get-all", methods=["GET"])
def get_all_intents(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, required_permissions=[],
        success_handler=_get_all_intents, to_json=True,
        project_url=project_url
    )

def _get_all_intents(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    intents_instances_list = project_resources.vocal_apps_model_schemas_dynamodb_client.get_all_intents_dicts_by_account_project_id(
        account_project_id=project_resources.account_project_id
    )
    return jsonify(success=True, intents=intents_instances_list)

@intents_blueprint.route("/<account_username>/<project_url>/intents/set/one", methods=["POST"])
def set_one_intent(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, required_permissions=[],
        success_handler=_set_one_intent, to_json=True,
        project_url=project_url
    )

def _set_one_intent(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.vocal_apps_model_schemas.models import InoftVocalEngineModelSchema
    try:
        intent_model = InoftVocalEngineModelSchema.IntentModel(**request.get_json())
        response = project_resources.vocal_apps_model_schemas_dynamodb_client.set_update_one_intent(
            accountProjectId=project_resources.project_owner_account_id, intent_model=intent_model
        )
    except ValidationError as e:
        print(f"Error while validating an intent model data : {e}")
        return jsonify(success=False)

@intents_blueprint.route("/<account_username>/<project_url>/intents/set/multiple", methods=["POST"])
def set_multiple_intents(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, required_permissions=[],
        success_handler=_set_multiple_intents, to_json=True,
        project_url=project_url
    )

def _set_multiple_intents(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.vocal_apps_model_schemas.models import InoftVocalEngineModelSchema
    updated_intents_ids: List[str] = list()
    # Instead of returning an error if any intent has a ValidationError or a database error, we keep a list of the ids
    # of all the intents that have been successfully updated, which we will returned to the client backend, so that it
    # will be able to know which intents have been updated and do not need to be send again, and which intents have had
    # an error, that need to be send back the next time the intents are saved. This approach, allow that if one of the
    # the intent send by the client has an error, it will not block all the potential others intents send by the client.
    intents_data_list = SafeDict(request.get_json()).get("intents").to_list(default=None)
    if intents_data_list is not None:
        for intent_data in intents_data_list:
            try:
                intent_model = InoftVocalEngineModelSchema.IntentModel(**intent_data)
                response = project_resources.vocal_apps_model_schemas_dynamodb_client.set_update_one_intent(
                    accountProjectId=project_resources.account_project_id, intent_model=intent_model
                )
                if response is not None:
                    # If the response is None, there has been an error.
                    updated_intents_ids.append(intent_model.intentId)
            except ValidationError as e:
                print(f"Error while validating an intent model data : {e}")

    return jsonify(updatedIntentsIds=updated_intents_ids)







def prompt():
    # Local imports, to make sure we do not import unnecessary stuff, unless the prompt function is called.
    import os
    import click
    from pathlib import Path
    import inoft_vocal_framework
    from inoft_vocal_framework.cli.cli_cache import CliCache

    def prompt_user_to_select_folderpath():
        return click.prompt(text="What is the root folder path of your project ? "
                                 "This is the default if you do not write anything :",
                            default=str(Path(os.path.dirname(os.path.realpath(inoft_vocal_framework.__file__))).parent))

    changed_root_folderpath = False
    app_project_root_folderpath = CliCache.cache().get("lastAppProjectRootFolderpath").to_str(default=None)

    if app_project_root_folderpath is None:
        app_project_root_folderpath = prompt_user_to_select_folderpath()
        changed_root_folderpath = True
    else:
        if not click.confirm(f"Do you want to deploy the project that is present in the following folder : {app_project_root_folderpath}"):
            app_project_root_folderpath = prompt_user_to_select_folderpath()
            changed_root_folderpath = True

    while not os.path.exists(app_project_root_folderpath):
        if click.confirm(text="The root folder path of the project has not been found."
                              "Do you want to select a new folderpath ? Otherwise the CLI will close."):
            app_project_root_folderpath = prompt_user_to_select_folderpath()
            changed_root_folderpath = True
        else:
            exit(200)

    if changed_root_folderpath is True:
        CliCache.cache().put("lastAppProjectRootFolderpath", app_project_root_folderpath)
        CliCache.save_cache_to_yaml()
        click.echo(f"Saved the folderpath of your project for {click.style(text='faster load next time', fg='blue')}")

    return app_project_root_folderpath



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2019 Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import struct
import logging

from math import ceil
from collections import namedtuple

log = logging.getLogger(__name__)

__all__ = ['RTPPacket', 'RTCPPacket', 'SilencePacket']


def decode(data):
    """Creates an :class:`RTPPacket` or an :class:`RTCPPacket`.

    Parameters
    -----------
    data : bytes
        The raw packet data.
    """

    # While technically unreliable, discord RTP packets (should)
    # always be distinguishable from RTCP packets.  RTCP packets
    # should always have 200-204 as their second byte, while RTP
    # packet are (probably) always 73 (or at least not 200-204).

    assert data[0] >> 6 == 2 # check version bits
    return _rtcp_map.get(data[1], RTPPacket)(data)

def is_rtcp(data):
    return 200 <= data[1] <= 204

def _parse_low(x):
    return x / 2.0 ** x.bit_length()


class _PacketCmpMixin:
    __slots__ = ()

    def __lt__(self, other):
        return self.timestamp < other.timestamp

    def __gt__(self, other):
        return self.timestamp > other.timestamp

    def __eq__(self, other):
        return self.timestamp == other.timestamp

class SilencePacket(_PacketCmpMixin):
    __slots__ = ('ssrc', 'timestamp')
    decrypted_data = b'\xF8\xFF\xFE'

    def __init__(self, ssrc, timestamp):
        self.ssrc = ssrc
        self.timestamp = timestamp

    def __repr__(self):
        return '<SilencePacket timestamp={0.timestamp}, ssrc={0.ssrc}>'.format(self)

class FECPacket(_PacketCmpMixin):
    __slots__ = ('ssrc', 'timestamp', 'sequence')
    decrypted_data = b''

    def __init__(self, ssrc, timestamp, sequence):
        self.ssrc = ssrc
        self.timestamp = sequence
        self.sequence = timestamp

    def __repr__(self):
        return '<FECPacket timestamp={0.timestamp}, sequence={0.sequence}, ssrc={0.ssrc}>'.format(self)

# Consider adding silence attribute to differentiate (to skip isinstance)

class RTPPacket(_PacketCmpMixin):
    __slots__ = ('version', 'padding', 'extended', 'cc', 'marker',
                 'payload', 'sequence', 'timestamp', 'ssrc', 'csrcs',
                 'header', 'data', 'decrypted_data', 'extension')

    _hstruct = struct.Struct('>xxHII')
    _ext_header = namedtuple("Extension", 'profile length values')

    def __init__(self, data):
        data = bytearray(data)

        self.version  =      data[0] >> 6
        self.padding  = bool(data[0] & 0b00100000)
        self.extended = bool(data[0] & 0b00010000)
        self.cc       =      data[0] & 0b00001111

        self.marker   = bool(data[1] & 0b10000000)
        self.payload  =      data[1] & 0b01111111

        self.sequence, self.timestamp, self.ssrc = self._hstruct.unpack_from(data)

        self.csrcs = ()
        self.extension = None

        self.header = data[:12]
        self.data = data[12:]
        self.decrypted_data = None

        if self.cc:
            fmt = '>%sI' % self.cc
            offset = struct.calcsize(fmt) + 12
            self.csrcs = struct.unpack(fmt, data[12:offset])
            self.data = data[offset:]

        # TODO?: impl padding calculations (though discord doesn't seem to use that bit)

    def update_ext_headers(self, data):
        """Adds extended header data to this packet, returns payload offset"""

        profile, length = struct.unpack_from('>HH', data)
        values = struct.unpack('>%sI' % length, data[4:4+length*4])
        self.extension = self._ext_header(profile, length, values)

        # TODO?: Update self.data with new data offset
        # ... (do I need to do this? because it seems to work fine without it)

        return 4 + length * 4

    def _dump_info(self):
        attrs = {name: getattr(self, name) for name in self.__slots__}
        return ''.join((
            "<RTPPacket ",
            *['{}={}, '.format(n, v) for n, v in attrs.items()],
            '>'))

    def __repr__(self):
        return '<RTPPacket ext={0.extended}, ' \
               'timestamp={0.timestamp}, sequence={0.sequence}, ' \
               'ssrc={0.ssrc}, size={1}' \
               '>'.format(self, len(self.data))

# http://www.rfcreader.com/#rfc3550_line855
class RTCPPacket(_PacketCmpMixin):
    __slots__ = ('version', 'padding', 'length')
    _header = struct.Struct('>BBH')
    _ssrc_fmt = struct.Struct('>I')
    type = None

    def __init__(self, data):
        head, _, self.length = self._header.unpack_from(data)
        self.version = head >> 6
        self.padding = bool(head & 0b00100000)
        # dubious, yet devious
        setattr(self, self.__slots__[0], head & 0b00011111)

    def __repr__(self):
        content = ', '.join("{}: {}".format(k, getattr(self, k, None)) for k in self.__slots__)
        return "<{} {}>".format(self.__class__.__name__, content)

    @classmethod
    def from_data(cls, data):
        _, ptype, _ = cls._header.unpack_from(data)
        return _rtcp_map[ptype](data)

# TODO?: consider moving repeated code to a ReportPacket type
# http://www.rfcreader.com/#rfc3550_line1614
class SenderReportPacket(RTCPPacket):
    __slots__ = ('report_count', 'ssrc', 'info', 'reports', 'extension')
    _info_fmt = struct.Struct('>5I')
    _report_fmt = struct.Struct('>IB3x4I')
    _24bit_int_fmt = struct.Struct('>4xI')
    _info = namedtuple('RRSenderInfo', 'ntp_ts rtp_ts packet_count octet_count')
    _report = namedtuple("RReport", 'ssrc perc_loss total_lost last_seq jitter lsr dlsr')
    type = 200

    def __init__(self, data):
        super().__init__(data)
        self.ssrc = self._ssrc_fmt.unpack_from(data, 4)[0]
        self.info = self._read_sender_info(data, 8)

        reports = []
        for x in range(self.report_count):
            offset = 28 + 24 * x
            reports.append(self._read_report(data, offset))

        self.reports = tuple(reports)

        self.extension = None
        if len(data) > 28 + 24*self.report_count:
            self.extension = data[28 + 24*self.report_count:]

    def _read_sender_info(self, data, offset):
        nhigh, nlow, rtp_ts, pcount, ocount = self._info_fmt.unpack_from(data, offset)
        ntotal = nhigh + _parse_low(nlow)
        return self._info(ntotal, rtp_ts, pcount, ocount)

    def _read_report(self, data, offset):
        ssrc, flost, seq, jit, lsr, dlsr = self._report_fmt.unpack_from(data, offset)
        clost = self._24bit_int_fmt.unpack_from(data, offset)[0] & 0xFFFFFF
        return self._report(ssrc, flost, clost, seq, jit, lsr, dlsr)

# http://www.rfcreader.com/#rfc3550_line1879
class ReceiverReportPacket(RTCPPacket):
    __slots__ = ('report_count', 'ssrc', 'reports', 'extension')
    _report_fmt = struct.Struct('>IB3x4I')
    _24bit_int_fmt = struct.Struct('>4xI')
    _report = namedtuple("RReport", 'ssrc perc_loss total_lost last_seq jitter lsr dlsr')
    type = 201

    def __init__(self, data):
        super().__init__(data)
        self.ssrc = self._ssrc_fmt.unpack_from(data, 4)[0]

        reports = []
        for x in range(self.report_count):
            offset = 8 + 24 * x
            reports.append(self._read_report(data, offset))

        self.reports = tuple(reports)

        self.extension = None
        if len(data) > 8 + 24*self.report_count:
            self.extension = data[8 + 24*self.report_count:]

    def _read_report(self, data, offset):
        ssrc, flost, seq, jit, lsr, dlsr = self._report_fmt.unpack_from(data, offset)
        clost = self._24bit_int_fmt.unpack_from(data, offset)[0] & 0xFFFFFF
        return self._report(ssrc, flost, clost, seq, jit, lsr, dlsr)

# UNFORTUNATELY it seems discord only uses the above ~~two packet types~~ packet type.
# Good thing I knew that when I made the rest of these. Haha yes.

# http://www.rfcreader.com/#rfc3550_line2024
class SDESPacket(RTCPPacket):
    __slots__ = ('source_count', 'chunks', '_pos')
    _item_header = struct.Struct('>BB')
    _chunk = namedtuple("SDESChunk", 'ssrc items')
    _item = namedtuple("SDESItem", 'type size length text')
    type = 202

    def __init__(self, data):
        super().__init__(data)
        _chunks = []
        self._pos = 4

        for x in range(self.source_count):
            _chunks.append(self._read_chunk(data))

        self.chunks = tuple(_chunks)

    def _read_chunk(self, data):
        ssrc = self._ssrc_fmt.unpack_from(data, self._pos)[0]
        self._pos += 4

        # check for chunk with no items
        if data[self._pos:self._pos+4] == b'\x00\x00\x00\x00':
            self._pos += 4
            return self._chunk(ssrc, ())

        items = [self._read_item(data)]

        # Read items until END type is found
        while items[-1].type != 0:
            items.append(self._read_item(data))

        # pad chunk to 4 bytes
        if self._pos % 4:
            self._pos = ceil(self._pos/4)*4

        return self._chunk(ssrc, items)

    def _read_item(self, data):
        itype, ilen = self._item_header.unpack_from(data, self._pos)
        self._pos += 2
        text = None

        if ilen:
            text = data[self._pos:self._pos+ilen].decode()
            self._pos += ilen

        return self._item(itype, ilen+2, ilen, text)

    def _get_chunk_size(self, chunk):
        return 4 + max(4, sum(i.size for i in chunk.items)) # + padding?

# http://www.rfcreader.com/#rfc3550_line2311
class BYEPacket(RTCPPacket):
    __slots__ = ('source_count', 'ssrcs', 'reason')
    type = 203

    def __init__(self, data):
        super().__init__(data)
        self.ssrcs = struct.unpack_from('>%sI' % self.source_count, data, 4)
        self.reason = None

        body_length = 4 + len(self.ssrcs) * 4
        if len(data) > body_length:
            extra_len = struct.unpack_from('B', data, body_length)[0]
            reason = struct.unpack_from('%ss' % extra_len, data, body_length + 1)
            self.reason = reason.decode()

# http://www.rfcreader.com/#rfc3550_line2353
class APPPacket(RTCPPacket):
    __slots__ = ('subtype', 'ssrc', 'name', 'data')
    _packet_info = struct.Struct('>I4s')
    type = 204

    def __init__(self, data):
        super().__init__(data)
        self.ssrc, name = self._packet_info.unpack_from(data, 4)
        self.name = name.decode('ascii')
        self.data = data[12:] # should be a multiple of 32 bits but idc

_rtcp_map = {
    200: SenderReportPacket,
    201: ReceiverReportPacket,
    202: SDESPacket,
    203: BYEPacket,
    204: APPPacket
}



from typing import List, Optional
from pydantic import BaseModel
from inoft_vocal_framework.inoft_vocal_markup.deserializer import DialogueLine


class ContentElement(BaseModel):
    id: str = None
    created_by: str = None
    created_on: str = None
    modified_on: str = None
    crude_text: str = None
    character_names: list = None
    dialogues_lines: Optional[List[DialogueLine]] = None



from typing import Optional, List

from pydantic import BaseModel


class DialogueLine(BaseModel):
    character_name: str = None
    line_content: str = None
    additional_character_metadata: Optional[str] = None

class ContentElement(BaseModel):
    id: str = None
    created_by: str = None
    created_on: str = None
    modified_on: str = None
    crude_text: str = None
    character_names: list = None
    dialogues_lines: Optional[List[DialogueLine]] = None

class Deserializer:
    def __init__(self, characters_names: Optional[list] = None):
        self.characters_names = characters_names

    def deserialize(self, text: str) -> List[DialogueLine]:
        lines = text.split("\n")
        dialogues_lines: List[DialogueLine] = list()

        for line in lines:
            current_dialogue_line = DialogueLine()

            has_found_first_star_in_line = False
            text_to_find_first_character_name = ""
            has_found_character_name = False

            for i_char, char in enumerate(line):
                if has_found_first_star_in_line is True:
                    text_to_find_first_character_name += char

                if has_found_character_name is False:
                    for character_name in self.characters_names:
                        if character_name in text_to_find_first_character_name:
                            current_dialogue_line.character_name = character_name
                            has_found_character_name = True

                    if char == "*":
                        has_found_first_star_in_line = True

                elif has_found_character_name is True:
                    if char == "*":
                        # Has found a star after having found a character name. Which we means, we have reached the end
                        # of the declaration of which character is talking, like the following : *Jean-Michel* Hi my friend !

                        # We remove the last character from the text to find the character name (so that we remove the current star char).
                        # And we remove the character name from the text, we strip of any additional junk chars, and if we still have
                        # something, it means that between the two stars where the character name has been found, there was more infos
                        # than just the character name, it is the additional character metadata that we add to the current dialogue line.
                        metadata = text_to_find_first_character_name[:-1].replace(current_dialogue_line.character_name, "").strip()
                        if metadata != "":
                            current_dialogue_line.additional_character_metadata = metadata

                        # Then we can use the rest of the line as what the user has said, and we add 1 the i_char,
                        # to not include the star char, and strip the text to remove any potential trailing spaces.
                        line_content = line[i_char + 1:].strip()
                        if line_content != "":
                            current_dialogue_line.line_content = line_content
                        break

                if i_char == len(line) - 1:
                    # If the char has is the last char of the line, and that the loop has not been
                    # break after finding a character name, then we will not set the character
                    # name, and just put the entire line as the dialogue line content.
                    line_content = line.strip()
                    if line_content != "":
                        current_dialogue_line.line_content = line_content

            if current_dialogue_line.line_content is not None:
                dialogues_lines.append(current_dialogue_line)
            elif current_dialogue_line.character_name is not None or current_dialogue_line.additional_character_metadata is not None:
                print(f"Warning ! A dialogue line was not empty of infos, but its content was empty : {current_dialogue_line.__dict__}")

        for line in dialogues_lines:
            print(line.__dict__)

        return dialogues_lines


if __name__ == "__main__":
    text_ = "*Léo parlant fort * Willie, t'arrête pas, ont y est presque !\n*bruits de course recommence et s'arrête*\n*Menu* Vous êtes arrivé au camp, vous êtes à une dizaine de mètre de la tente de communication. Une voiture KubelWagen couleur sable se trouve à votre droite. Une grande roue est accroché sur le capeau, et une mitrailleuse est monté au milieu de la voiture. Elle est suffisamment haute pour que que vous puissiez vous cacher derrière ou même en dessous de leurs uniformes."
    Deserializer(characters_names=["Léo", "Willie", "Menu"]).deserialize(text=text_)


from collections import Callable

import click
from discord import Message


def start_discord_listening(token: str, lambda_handler_function: Callable, command_prefix: str = "."):
    if click.confirm("Are you sure you want to trigger a Discord listening event ?"
                     "Never do that on a serverless app, since the app will keep running until closed."):
        DiscordHandlerInput.SHOULD_BE_USED = True
        DiscordHandlerInput.TOKEN = token
        DiscordHandlerInput.HANDLER_FUNCTION_TO_USE = lambda_handler_function

        from inoft_vocal_framework.platforms_handlers.discord import ClientWrapper
        DiscordHandlerInput.CLIENT_WRAPPER = ClientWrapper(prefix=command_prefix)
        DiscordHandlerInput.CLIENT_WRAPPER.run(token=token)


class DiscordHandlerInput:
    from inoft_vocal_framework.platforms_handlers.handler_input import HandlerInput

    SHOULD_BE_USED = False
    TOKEN = None
    HANDLER_FUNCTION_TO_USE = None
    CLIENT_WRAPPER = None

    def __init__(self, parent_handler_input: HandlerInput, request: Message):
        self.parent_handler_input = parent_handler_input
        self.request = request

    async def _say(self, text_or_ssml: str):
        data = {
            "author_name": self.request.author.name,
            "author_id": self.request.author.id,
            "message": text_or_ssml,
            "time": self.request.created_at
        }
        print(data)
        # public channel sending
        await self.request.channel.send(text_or_ssml)

    def say(self, text_or_ssml: str):
        import asyncio
        loop = asyncio.get_event_loop()
        loop.create_task(self._say(text_or_ssml=text_or_ssml))
        loop.run_until_complete()
        # todo: understand why this line crash, and why when this line is on, only the first message will be send




from typing import Optional

import click

from inoft_vocal_framework.dummy_object import dummy_object
from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type_and_not_none
from inoft_vocal_framework.skill_settings.settings_components.deployment import Deployment
from inoft_vocal_framework.skill_settings.settings_components.dynamodb_databases import DatabaseSessionsUsersData, \
    DatabaseMessagesContent, DatabaseUsersNotificationsSubscriptions
from inoft_vocal_framework.skill_settings.settings_components.plugins import Plugins
# from inoft_vocal_engine.speech_synthesis.polly import VOICES
# todo: currently, the voices setting is deprecated since the split between the inoft_vocal_framework and the inoft_vocal_engine


def prompt_database_warning_message(variable_name: str, instance_type: type):
    click.echo(click.style(f"\nWarning ! The variable {variable_name} for the {instance_type.__name__} instance was not properly set.\n"
                           f"If you do not need this database client, please set the disable_database variable to True on this instance.",
                           fg="yellow"))


class Settings:
    def __init__(self, characters_voices: Optional[dict] = None, deployment: Optional[Deployment] = None,
                 database_sessions_users_data: DatabaseSessionsUsersData = DatabaseSessionsUsersData(),
                 database_messages_content: DatabaseMessagesContent = DatabaseMessagesContent(),
                 database_users_notifications_subscriptions: DatabaseUsersNotificationsSubscriptions = DatabaseUsersNotificationsSubscriptions(),
                 plugins: Plugins = Plugins()):
        self.characters_voices = characters_voices
        self.deployment = deployment
        self.database_sessions_users_data = database_sessions_users_data
        self.database_messages_content = database_messages_content if database_messages_content is not None else dummy_object
        self.database_users_notifications_subscriptions = database_users_notifications_subscriptions if database_users_notifications_subscriptions is not None else dummy_object
        self.default_session_data_timeout = 60
        self.plugins = plugins if plugins is not None else list()  # We use a list instead of dummy object, since there is multiple for loop calls on this object

    @property
    def characters_voices(self) -> dict:
        return self._characters_voices if self._characters_voices is not None else dummy_object

    @characters_voices.setter
    def characters_voices(self, characters_voices: characters_voices) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=characters_voices, expected_type=dict, variable_name="characters_voices")
        self._characters_voices = characters_voices

    @property
    def deployment(self) -> Deployment:
        return self._deployment if self._deployment is not None else dummy_object

    @deployment.setter
    def deployment(self, deployment: Deployment) -> None:
        raise_if_variable_not_expected_type_and_not_none(value=deployment, expected_type=Deployment, variable_name="deployment")
        self._deployment = deployment


def prompt_get_settings(root_folderpath: Optional[str] = None) -> Settings:
    import os

    if root_folderpath is None:
        from inoft_vocal_framework.cli.components import current_project_directory
        root_folderpath = current_project_directory.prompt()

    print("Searching for your app_settings file...")

    found_settings_filepaths = []
    for filename in os.listdir(root_folderpath):
        if filename in ["app_settings.py", "settings.py"]:
            found_settings_filepaths.append(os.path.join(root_folderpath, filename))

    settings_filepath = None
    if len(found_settings_filepaths) == 0:
        raise Exception("Did not found an app_settings.py or settings.py file.\nDid you renamed your settings file ?"
                        "\nOr did you not navigated to the directory of your project ?"
                        "\nTry to run the command cd 'C:/folder/subfolder/myproject'."
                        "\nIf your project is on an external hard-drive (like F:), you must run the cd command, and also the letter name of your hard-drive."
                        "\nSo if your project is in F:/folder/subfolder/myproject, in your command line, run   F:   then run   cd F:/folder/subfolder/myproject")
    elif len(found_settings_filepaths) == 1:
        print(f"Using setting file {found_settings_filepaths[0]}")
        settings_filepath = found_settings_filepaths[0]
    elif len(found_settings_filepaths) > 1:
        while settings_filepath is None:
            print(f"Found {len(found_settings_filepaths)} settings file at following paths :")
            for i_file, file in enumerate(found_settings_filepaths):
                print(f"{i_file + 1} - {file}")

            import click
            file_index = str(click.prompt("What is the number of the file you wish to use ?"))
            if not file_index.isdigit() or int(file_index) > len(found_settings_filepaths) or int(file_index) < 0:
                print("Please write a valid file index in the form of an int like 1 or 3")
            else:
                selected_filepath = found_settings_filepaths[int(file_index) - 1]
                print(f"Selected settings filepath : {click.style(text=selected_filepath, fg='blue')}")
                settings_filepath = selected_filepath


    import importlib.util
    spec = importlib.util.spec_from_file_location("plugin_core", settings_filepath)
    current_plugin_core_module = importlib.util.module_from_spec(spec=spec)
    spec.loader.exec_module(current_plugin_core_module)

    vars_current_settings_module = vars(current_plugin_core_module)
    if "settings" not in vars_current_settings_module.keys():
        raise Exception(f"The variable named 'settings' is missing from the settings file you selected at {settings_filepath}")
    else:
        settings_instance = vars_current_settings_module["settings"]
        if isinstance(settings_instance, Settings):
            return settings_instance
        else:
            raise Exception(f"A variable named settings has been found in the file {settings_instance}, but it was not an instance of Settings type.")
    # This code line is unreachable, no need for an additional exception


if __name__ == "__main__":
    Settings(characters_voices={})
    """
    "Léo": VOICES.French_France_Male_MATHIEU,
    "Willie": VOICES.French_France_Female_CELINE,
    "Luc": VOICES.Russian_Russia_Male_MAXIM,
    "Menu": VOICES.Icelandic_Iceland_Male_KARL,
    "default": VOICES.French_France_Female_CELINE
    """



from flask.helpers import url_for


def dynamic_url_for(endpoint, **values):
    rv = url_for(endpoint=endpoint, **values)
    from flask import current_app as app
    if app.config["shouldUseCloudDist"] is True:
        rv = f"{app.config['bucketUrl']}/{app.config['bucketDirPath']}{rv}"
    return rv



from pydantic import BaseModel


class BaseCondition(BaseModel):
    conditionType: str

class ConditionIsIntentName(BaseCondition):
    _TYPE = "IS-INTENT-NAME"
    intentName: str

    def __init__(self, **kwargs):
        super().__init__(conditionType=self._TYPE, **kwargs)

class ConditionAlways(BaseCondition):
    _TYPE = "ALWAYS"

    def __init__(self, **kwargs):
        super().__init__(conditionType=self._TYPE, **kwargs)

class ConditionMemoryVariableExpression(BaseCondition):
    _TYPE = "MEMORY-VARIABLE-EXPRESSION"
    memoryType: str
    variableName: str
    conditionExpression: str

    def __init__(self, **kwargs):
        super().__init__(conditionType=self._TYPE, **kwargs)

def get_appropriate_condition(condition_data: dict):
    # todo: improve this code
    if isinstance(condition_data, dict) and "conditionType" in condition_data.keys():
        condition_type = condition_data["conditionType"]
        condition_data.pop("conditionType")

        if condition_type == ConditionIsIntentName._TYPE:
            return ConditionIsIntentName(**condition_data)
        elif condition_type == ConditionAlways._TYPE:
            return ConditionAlways(**condition_data)
        elif condition_type == ConditionMemoryVariableExpression._TYPE:
            return ConditionMemoryVariableExpression(**condition_data)
    return None



import ast
import os
from typing import Dict, List
import inflect as inflect

from inoft_vocal_engine.code_generation.models import MessageItem, SayAction, SetVariableAction
from inoft_vocal_engine.import_integrations.botpress.engine_models import ProcessedNodeElement, DatabaseTransitionConditionItem
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.utils.general import load_json
from inoft_vocal_engine.code_generation.templates.templates_access import TemplatesAccess

# todo: fix bug with node where no text is played (it will jump the node itself, and the next node x)  )


class GeneratorCore:
    def __init__(self, nodes_list: List[ProcessedNodeElement], main_flow_filepath: str = None, builtin_text_filepath: str = None):
        # self.main_flow_filepath = main_flow_filepath
        # self.builtin_text_filepath = builtin_text_filepath
        self.nodes_list = nodes_list

        self.has_conditions_classes = False
        self.has_request_handlers = False
        self.has_state_handlers = False
        self.messages = None

        self.threshold_of_intent_use_to_create_a_condition = 2

        self.counts_used_condition_intent_names = dict()
        self.node_ids_to_data_instances: Dict[str, ProcessedNodeElement] = dict()
        self.node_ids_to_components_instances = dict()

        self.templates = TemplatesAccess()
        self._plugins = None
        self.plugins_folder_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "plugins")

        self.say_actions: List[SayAction] = list()
        self.set_variables_actions: List[SetVariableAction] = list()

    @property
    def plugins(self) -> list:
        if self._plugins is None:
            self.plugins_folder_dir = os.path.dirname(os.path.abspath(__file__))
        return list()

    def generate(self):
        from inoft_vocal_engine.code_generation.components.launch_request_handler import LaunchRequestHandler
        from inoft_vocal_engine.code_generation.components.state_handler import StateHandler
        from inoft_vocal_engine.code_generation.components.intent_name_condition import IntentNameCondition
        from inoft_vocal_engine.code_generation.components.messages import Messages

        # self.messages = Messages(messages_items=load_json(self.builtin_text_filepath))
        # self.write_to_file(text=self.messages.render(parent_core=self), filepath="F:/Inoft/anvers_1944_project/inoft_vocal_engine/dist/messages.py")
        # os.path.join(os.path.dirname(os.path.abspath(__file__)), "messages.py"))

        # todo: re-implement logic of the startNode
        """
        flow_dict = load_json(filepath=self.main_flow_filepath)
        if "nodes" not in flow_dict.keys():
            raise Exception(f"The nodes key has not been found in the dict of the flow_file : {flow_dict}")
        else:
            name_start_node = flow_dict["startNode"]
            nodes_list = flow_dict["nodes"]
            # Initialization of the classes
        """

        for node in self.nodes_list:
            self.node_ids_to_data_instances[node.node_name] = node

            """if node_name == name_start_node:
                self.node_ids_to_components_instances[node_name] = LaunchRequestHandler(node_dict=node)
                self.has_request_handlers = True
            else:"""
            self.node_ids_to_components_instances[node.node_name] = StateHandler(node_name=node.node_name)
            self.has_state_handlers = True

        # Processing of the classes and their interactions with each others
        for node_class in self.node_ids_to_components_instances.values():
            node_class.process(parent_core=self)

            if isinstance(node_class, StateHandler):
                for intent_key, count_value in node_class.counts_used_condition_intent_names.items():
                    if intent_key in self.counts_used_condition_intent_names.keys():
                        self.counts_used_condition_intent_names[intent_key] += count_value
                    else:
                        self.counts_used_condition_intent_names[intent_key] = count_value

        output_conditions_classes = list()
        for intent_key, count_value in self.counts_used_condition_intent_names.items():
            if count_value >= self.threshold_of_intent_use_to_create_a_condition:
                new_intent_name_condition_class = IntentNameCondition(intent_name=intent_key)
                new_intent_name_condition_class.render(parent_core=self)
                output_conditions_classes.append(new_intent_name_condition_class)
                self.has_conditions_classes = True

        output_handlers_list = list()
        for class_from_node in self.node_ids_to_components_instances.values():
            for handler_class in class_from_node.render(parent_core=self):
                output_handlers_list.append(handler_class)

        skill_app_rendered_code = self.templates.skill_app_template.render(conditions_classes_list=output_conditions_classes,
            handlers_list=output_handlers_list, has_condition_classes=self.has_conditions_classes,
            has_request_handlers=self.has_request_handlers, has_state_handlers=self.has_state_handlers)

        return skill_app_rendered_code

        # self.write_to_file(text=skill_app_rendered_code, filepath="F:/Inoft/anvers_1944_project/inoft_vocal_engine/dist/app_generated.py")

    def process_on_enter(self, actions_list: list) -> list:
        self.say_actions.clear()
        logic_elements = list()

        for action in actions_list:
            if isinstance(action, str):
                if len(action) >= 3:
                    if action[0:3] == "say":
                        if "#!builtin_text" in action:
                            current_say_action = SayAction()
                            current_say_action.message_id_to_say = action.replace("say", "").replace("#!", "").replace(" ", "")

                            if current_say_action.message_id_to_say in self.messages.output_messages_dict.keys():
                                current_say_action.message_item_to_say = self.messages.output_messages_dict[current_say_action.message_id_to_say]
                            else:
                                current_say_action.text_to_say_if_message_item_missing = f"The message with id {current_say_action.message_id_to_say} is missing"

                            self.say_actions.append(current_say_action)

                if "/setVariable" in action:
                    processed_action_list = action.split('{', maxsplit=1)
                    if len(processed_action_list) > 1:
                        processed_action_dict = ast.literal_eval("{" + processed_action_list[1].replace("}", "") + "}")
                        if isinstance(processed_action_dict, dict):
                            if all(key in processed_action_dict.keys() and processed_action_dict[key] != "" for key in ["name", "value"]):
                                if "type" in processed_action_dict.keys() and processed_action_dict["type"] == "str":
                                    processed_action_dict["value"] = f'"{processed_action_dict["value"]}"'

                                logic_elements.append(self.templates.set_variable_logic_template.render(action_dict=processed_action_dict))

        # from inoft_vocal_framework.plugins.OfficialAudioFilesInsteadOfText import Core as PluginCore
        # PluginCore().execute(generator_core=self)
        # todo: make it possible for the engine to use plugins, and create a new plugins system for the framework,
        #  that will not affect the code generation, but give the ability to add new functions to the framework.

        for say_action in self.say_actions:
            if say_action.code is None:
                if say_action.message_item_to_say is not None:
                    say_action.code = f"{say_action.message_item_to_say.variable_name}.pick()"
                    say_action.is_callable = True
                else:
                    say_action.code = say_action.text_to_say_if_message_item_missing
                    say_action.is_callable = False

            logic_elements.append(self.templates.say_action_template.render(say_action=say_action, **say_action.extra_args))

        return logic_elements

    @staticmethod
    def write_to_file(text: str, filepath: str):
        with open(filepath, "w+", encoding="utf-8") as file:
            file.write(text)

if __name__ == "__main__":
    GeneratorCore(nodes_list=list(),
        main_flow_filepath="F:/Inoft/anvers_1944_project/inoft_vocal_engine/code_generation/main.flow.json",
        builtin_text_filepath="F:/Inoft/anvers_1944_project/inoft_vocal_engine/code_generation/builtin_text.json"
                  ).generate()
    print("done")



from abc import abstractmethod
from typing import Any


class PluginBase:
    # Exist to make it easier to detect if a class if a plugin of any class type.
    pass


class PluginCodeGenerationBase(PluginBase):
    # from inoft_vocal_framework.botpress_integration.generator import GeneratorCore

    @abstractmethod
    def execute(self, generator_core: Any):  # GeneratorCore):
        raise Exception("Execute method of the plugin must be implemented. "
                        "It will be the method that will be run when your plugin will be execute. "
                        "Also, all the classical class methods (like __init__) will be run like usual.")



import os
from inoft_vocal_engine import InoftSkill, InoftRequestHandler, InoftStateHandler, InoftDefaultFallback


class StartRequestHandler(InoftRequestHandler):
    def can_handle(self) -> bool:
        return self.is_launch_request()

    def handle(self) -> dict:
        if not self.google.has_user_granted_push_notifications_permission():
            self.say("Ca vous dirait que je vous envoie une notification quand le prochain chapitre sort ?")
            self.memorize_session_then_state(OfferNotificationsSubscriptionStateHandler)
        else:
            self.say("Vous êtes déjà abonner aux notifications pour le prochain chapitre ! Aller salut.")
            self.end_session()

        return self.to_platform_dict()


class OfferNotificationsSubscriptionStateHandler(InoftStateHandler):
    def handle(self) -> dict:
        if self.is_in_intent_names("Yes"):
            self.say("Très bien ! Google Assistant vas vous demander l'autorisation de vous envoyer les notifications.")
            self.google.request_push_notifications_permission_if_missing(intent_name="actions.intent.MAIN")
            self.memorize_session_then_state(NotificationsSubscriptionsCallbacksStateHandler)

        elif self.is_in_intent_names("No"):
            self.say("Vous êtes vraiment méchant ! Je boude maintenant, aurevoir.")
            self.end_session()

        return self.to_platform_dict()

    def fallback(self) -> dict:
        self.say("Alors, vous voulez que je vous inscrive aux notifications ? Oui, non ?")
        return self.to_platform_dict()


class NotificationsSubscriptionsCallbacksStateHandler(InoftStateHandler):
    def handle(self) -> dict:
        if self.google.has_user_granted_push_notifications_permission():
            self.google.subscribe_current_user_to_notifications_group("nextChapterRelease")
            self.say("Super, je vous ai abonné ! Excellente journée à vous !")
            self.end_session()
        else:
            self.say("C'est vraiment pas sympa de ne pas avoir accepté les notifications ! Je vous fais la tête, aller salut.")
            self.end_session()

        return self.to_platform_dict()


class DefaultFallback(InoftDefaultFallback):
    def handle(self):
        self.say("J'ai aucune idée de ce que vous voulez dire. Aller salut.")
        self.end_session()
        return self.to_platform_dict()


def lambda_handler(event, context):
    skill = InoftSkill()
    skill.add_request_handler(StartRequestHandler)
    skill.add_state_handler(OfferNotificationsSubscriptionStateHandler)
    skill.add_state_handler(NotificationsSubscriptionsCallbacksStateHandler)
    skill.set_default_fallback_handler(DefaultFallback)
    return skill.handle_any_platform(event=event, context=context)

if __name__ == "__main__":
    from inoft_vocal_engine import Simulator
    event_, context_ = Simulator(platform=Simulator.PLATFORM_GOOGLE, event_type="launch").get_event_and_context()
    # event_, context_ = Simulator(platform=Simulator.PLATFORM_GOOGLE, event_type="grant_update_permissions").get_event_and_context()
    print(f"\n\nFinal Output : {lambda_handler(event=event_, context=context_)}")



from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type


class CurrentUsedPlatformInfo:
    def __init__(self):
        self._is_alexa = False
        self._is_dialogflow = False
        self._is_bixby = False
        self._is_discord = False

    @property
    def is_alexa(self) -> bool:
        return self._is_alexa

    @is_alexa.setter
    def is_alexa(self, is_alexa: bool) -> None:
        raise_if_variable_not_expected_type(value=is_alexa, expected_type=bool, variable_name="is_alexa")
        self._is_alexa = is_alexa

    @property
    def is_dialogflow(self) -> bool:
        return self._is_dialogflow

    @is_dialogflow.setter
    def is_dialogflow(self, is_dialogflow: bool) -> None:
        raise_if_variable_not_expected_type(value=is_dialogflow, expected_type=bool, variable_name="is_dialogflow")
        self._is_dialogflow = is_dialogflow

    @property
    def is_bixby(self) -> bool:
        return self._is_bixby

    @is_bixby.setter
    def is_bixby(self, is_bixby: bool) -> None:
        raise_if_variable_not_expected_type(value=is_bixby, expected_type=bool, variable_name="is_bixby")
        self._is_bixby = is_bixby

    @property
    def is_discord(self) -> bool:
        return self._is_discord

    @is_discord.setter
    def is_discord(self, is_discord: bool) -> None:
        raise_if_variable_not_expected_type(value=is_discord, expected_type=bool, variable_name="is_discord")
        self._is_discord = is_discord



import os

from jinja2 import Template, FileSystemLoader, Environment
from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path


class TemplatesAccess:
    file_loader = FileSystemLoader(os.path.join(get_inoft_vocal_engine_root_path(), "code_generation/templates"))
    # We cannot just set the templates folder name in the FileSystemLoader, otherwise jinja will try to descend into
    # all the folders it found until it finds a folder called templates. And of course, the engine has other folders
    # named templates (in particular, the templates folder in the web_interface) folder. So, we need to use a direct
    # path to the code_generation/templates folder.
    env = Environment(loader=file_loader, lstrip_blocks=True, trim_blocks=True)

    # todo: remove the staticness

    _skill_app_template = None
    _launch_request_handler_template = None
    _intent_name_condition_template = None
    _state_handler_template = None
    _say_action_template = None
    _set_variable_logic_template = None
    _handler_logic_template = None
    _messages_template = None

    @staticmethod
    def _get_template(template_name: str) -> Template:
        return TemplatesAccess.env.get_template(template_name)

    @staticmethod
    def _load_template(template_filepath: str) -> Template:
        from inoft_vocal_engine.code_generation.templates.jinja_filepath_loader import JinjaFilepathLoader
        env = Environment(loader=JinjaFilepathLoader(filepath=template_filepath))
        # The JinjaFilepathLoader do not care about the template name, only the full template_filepath.
        return env.get_template(name=None)

    @property
    def skill_app_template(self) -> Template:
        if TemplatesAccess._skill_app_template is None:
            TemplatesAccess._skill_app_template = TemplatesAccess._get_template("skill_app.tem")
        return TemplatesAccess._skill_app_template

    @property
    def launch_request_handler_template(self) -> Template:
        if TemplatesAccess._launch_request_handler_template is None:
            TemplatesAccess._launch_request_handler_template = TemplatesAccess._get_template("launch_request_handler.tem")
        return TemplatesAccess._launch_request_handler_template

    @property
    def intent_name_condition_template(self) -> Template:
        if TemplatesAccess._intent_name_condition_template is None:
            TemplatesAccess._intent_name_condition_template = TemplatesAccess._get_template("intent_name_condition.tem")
        return TemplatesAccess._intent_name_condition_template

    @property
    def state_handler_template(self) -> Template:
        if TemplatesAccess._state_handler_template is None:
            TemplatesAccess._state_handler_template = TemplatesAccess._get_template("state_handler.tem")
        return TemplatesAccess._state_handler_template

    @property
    def say_action_template(self) -> Template:
        if TemplatesAccess._say_action_template is None:
            TemplatesAccess._say_action_template = TemplatesAccess._get_template("say_action.tem")
        return TemplatesAccess._say_action_template

    @say_action_template.setter
    def say_action_template(self, filepath: str):
        TemplatesAccess._say_action_template = self._load_template(template_filepath=filepath)

    @property
    def set_variable_logic_template(self) -> Template:
        if TemplatesAccess._set_variable_logic_template is None:
            TemplatesAccess._set_variable_logic_template = TemplatesAccess._get_template("set_variable_logic.tem")
        return TemplatesAccess._set_variable_logic_template

    @property
    def handler_logic_template(self) -> Template:
        if TemplatesAccess._handler_logic_template is None:
            TemplatesAccess._handler_logic_template = TemplatesAccess._get_template("handler_logic.tem")
        return TemplatesAccess._handler_logic_template

    @property
    def messages_template(self) -> Template:
        if TemplatesAccess._messages_template is None:
            TemplatesAccess._messages_template = TemplatesAccess._get_template("messages.tem")
        return TemplatesAccess._messages_template

    @messages_template.setter
    def messages_template(self, filepath: str):
        TemplatesAccess._messages_template = self._load_template(template_filepath=filepath)




from flask import Blueprint, render_template, jsonify, request

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow
from inoft_vocal_engine.web_interface.route_blueprints.audio_project import AudioProject
from inoft_vocal_engine.web_interface.static.ui import SIDEBAR_BUTTONS
from inoft_vocal_engine.web_interface.templates.topbar import topbar_ui

audio_editor_blueprint = Blueprint("audio-editor", __name__, template_folder="templates")

dice = {
    'projectId': True,
    'collections': {
        'tracks': {
            'models': [{
                'attributes': {
                    'file': {
                        'name': 'jean sablon - alexa.mp3',
                        'webkitRelativePath': '',
                        'lastModified': 1591264478848.0,
                        'size': 1007568.0,
                        'type': 'audio/mpeg'
                    },
                    'color': '#00a0b0',
                    'length': 1920.0,
                    'name': 'Track 1',
                    'solo': False,
                    'buffer': {
                        'duration': 167.71,
                        'length': 7396188.0,
                        'sampleRate': 44100.0,
                        'numberOfChannels': 2.0
                    },
                    'pan': 0,
                    'muted': False,
                    'gain': 1.0,
                }
            }]
        },
    }
}

if __name__ == "__main__":
    audio_project = AudioProject(**dice)
    audio_project.add_track()
    print(audio_project)


@audio_editor_blueprint.route("/audio-editor-v2")
def audio_editor_v2():
    return render_template("audio-editor-v2/index.html")

@audio_editor_blueprint.route("/<account_username>/<project_url>/audio-editor")
def audio_editor(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_audio_editor, to_template=True,
    )


def _audio_editor(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    # audio_project_data, audio_request_success = project_resources.audio_editor_projects_dynamodb_client.get_project_data_by_account_project_id(account_project_id=project_resources.account_project_id)
    # text_project_data, text_request_success = project_resources.project_text_contents_dynamodb_client.get_by_id(account_project_id=project_resources.account_project_id)
    # audio_project = NestedObjectToDict.dict_to_typed_class(class_type=AudioProject, data_dict=audio_project_data)
    # print(f"audio_project = {audio_project}")
    # print(f"audio_project_data = {audio_project_data}\n& text_project_data = {text_project_data}")

    static_data = {'Collections': {'Tracks': {'models': [{'attributes': {
        'buffer': {'duration': 167.714, 'length': 8050272, 'numberOfChannels': 2, 'sampleRate': 48000},
        'color': '#00a0b0', 'file': {'lastModified': 1591264478848, 'name': 'jean sablon - alexa.mp3', 'size': 1007568,
                                     'type': 'audio/mpeg', 'webkitRelativePath': ''}, 'gain': 1, 'length': 1920,
        'muted': False, 'name': 'Track 1', 'pan': 0.5, 'solo': False}}]}}}

    return render_template("audio-editor/index.html", project_data=static_data,
                           sidebar_buttons=SIDEBAR_BUTTONS,
                           topbar_buttons=topbar_ui.render(
                               account_username=project_resources.project_owner_account_username,
                               project_url=project_resources.project_url,
                               active_button_id=topbar_ui.ID_ORGANIZATION
                           ), **project_resources.required_template_kwargs())

@audio_editor_blueprint.route("/<account_username>/<project_url>/audio-editor/save", methods=["POST"])
def audio_editor_save(account_username: str, project_url: str):
    project_resources = ProjectResources(engine_resources=EngineResources(), account_username=account_username, project_url=project_url)
    print(f"Saving data for project {project_url}")

    request_json_data = request.get_json()
    if request_json_data is not None:
        request_success = project_resources.audio_editor_projects_dynamodb_client.save_project_data(
            account_project_id=project_resources.account_project_id, project_data=request_json_data)
    else:
        request_success = False

    return jsonify({"success": request_success})



import os
import time
from typing import List

from inoft_vocal_engine.audacity.client import AudacityClient
from inoft_vocal_engine.speech_synthesis.polly.client import PollyClient
from inoft_vocal_engine.speech_synthesis.polly import VOICES
from inoft_vocal_engine.inoft_vocal_markup.deserializer import Deserializer, DialogueLine
from inoft_vocal_engine import Sound
from inoft_vocal_engine.speech_synthesis.polly.voice_object import Voice


class TextToAudacity:
    def __init__(self, character_names_to_voices: dict):
        self._audacity = None
        self._polly = None
        self.character_names_to_voices = character_names_to_voices
        self.available_character_names = list(character_names_to_voices.keys())
        self.dialogue_deserializer = Deserializer(characters_names=self.available_character_names)

    @property
    def audacity(self) -> AudacityClient:
        if self._audacity is None:
            self._audacity = AudacityClient()
        return self._audacity

    @property
    def polly(self) -> PollyClient:
        if self._polly is None:
            self._polly = PollyClient()
        return self._polly

    def synthesize_text_to_project(self, text: str):
        dialogues_lines = self.dialogue_deserializer.deserialize(text=text)
        return self.synthesize_dialogues_lines_to_project(dialogues_lines=dialogues_lines)

    def synthesize_dialogues_lines_to_project(self, dialogues_lines: List[DialogueLine], dirpath_to_save_to: str):
        synthesized_dialogues: List[Sound] = list()

        element_lines_dirpath = os.path.join(dirpath_to_save_to, "lines")
        if not os.path.isdir(element_lines_dirpath):
            os.makedirs(element_lines_dirpath)

        for i_dialogue, dialogue in enumerate(dialogues_lines):
            if dialogue.character_name not in self.available_character_names:
                if "default" not in self.available_character_names:
                    raise Exception(f"The character name {dialogue.character_name} has not been found in the available character"
                                    f" names : {self.available_character_names} and no character named default has been found .")
                else:
                    dialogue.character_name = "default"

            voice_for_current_dialogue: Voice = self.character_names_to_voices[dialogue.character_name]
            if not isinstance(voice_for_current_dialogue, Voice):
                raise Exception(f"The voice object for the character name {dialogue.character_name} was "
                                f"not of type Voice but was of type : {type(voice_for_current_dialogue)}")

            from text_unidecode import unidecode
            character_name_without_accents = unidecode(dialogue.character_name)
            current_dialogue_line_filename = f"line_{str(i_dialogue).zfill(3)}__{character_name_without_accents}_{voice_for_current_dialogue.id}.mp3"

            filepath_to_save_to = os.path.join(element_lines_dirpath, current_dialogue_line_filename)
            if os.path.exists(filepath_to_save_to):
                current_synthesized_dialogue = Sound(filepath_to_save_to)
                synthesized_dialogues.append(current_synthesized_dialogue)
                print(f"Find and will be using existing file at {filepath_to_save_to}")
            else:
                synthesized_dialogue_filepath = self.polly.synthesize(text=dialogue.line_content,
                    voice_id=voice_for_current_dialogue.id, filepath_to_save_to=filepath_to_save_to)
                print(f"Synthesized new audio file to {filepath_to_save_to}")

                if synthesized_dialogue_filepath is not None:
                    current_synthesized_dialogue = Sound(synthesized_dialogue_filepath)
                    synthesized_dialogues.append(current_synthesized_dialogue)

        if len(synthesized_dialogues) > 0:
            self.audacity.delete_all_tracks()
            project_seconds_duration = 0.0

            for i_audio_dialogue, audio_dialogue in enumerate(synthesized_dialogues):
                self.audacity.import_file(audio_dialogue.local_filepath, track_number=i_audio_dialogue)

                if i_audio_dialogue > 0:
                    self.audacity.set_clip(clip_id=i_audio_dialogue, track_number=0, seconds_start=project_seconds_duration)
                project_seconds_duration += audio_dialogue.duration_seconds

            # self.audacity.mix_and_render_tracks()



if __name__ == "__main__":
    client = TextToAudacity(character_names_to_voices={
        "Léo": VOICES.French_France_Male_MATHIEU,
        "Willie": VOICES.French_France_Female_CELINE,
        "Menu": VOICES.Icelandic_Iceland_Male_KARL
    })

    text_ = """
    *Léo* Je voudrais tuer tous les allemands !\n
    *Willie* Comment on fait ? Il faudrait qu'ils croient qu'on est dans leur camp ?\n
    *Léo* Je ne sais pas exactement, on trouvera. Ont pourrait essayer de capturer un Boche ou deux.\n
    *Willie* Oui, ils sont très très très très moches les allemands, il faut tousse les tuer...
    """

    client.synthesize_text_to_project(text=text_)
    input()

    """
    leo = Sound(polly.synthesize(text="Ont pourrais essayer de récuperer des informations en utilisant leurs radio.",
                           voice_id=VOICES.French_France_Male_MATHIEU.id,
                                 filepath_to_save_to="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/leo.mp3"))

    willie = Sound(polly.synthesize(text="Comment ont ferrais, il faudrais qu'ils croient qu'ont soit dans leurs camp ?",
                                    voice_id=VOICES.French_France_Female_CELINE.id,
                                    filepath_to_save_to="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/willie.mp3"))
    """

    leo = Sound("F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/leo.mp3")
    willie = Sound("F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/willie.mp3")

    audacity.delete_all_tracks()
    audacity.import_file(leo.local_filepath, track_number=1)
    # audacity.import_file(willie.local_filepath, track_number=2)
    audacity.set_clip(clip_id=1, track_number=0, seconds_start=leo.duration_seconds)



import io
import json

import requests
from google.oauth2 import service_account
import google.auth.transport.requests

PATH_TO_SERVICE_ACCOUNT = "C:/Users/LABOURDETTE/Downloads/jeu-seconde-guerre-mondiale-84a61cf2b3ca.json"

REQUIRED_SCOPE = "https://www.googleapis.com/auth/actions.fulfillment.conversation"

# Get access token
with io.open(PATH_TO_SERVICE_ACCOUNT, "r", encoding="utf-8") as json_fi:
    credentials_info = json.load(json_fi)

credentials = service_account.Credentials.from_service_account_info(credentials_info, scopes=[REQUIRED_SCOPE])
request = google.auth.transport.requests.Request()
credentials.refresh(request)

headers = {
    "Authorization": "Bearer " + credentials.token
}

payload = {
    "customPushMessage": {
        "userNotification": {
            "title": "Notification title",
            "text": "Simple Text"
        },
        "target": {
            "userId": "<USER_ID>",
            "intent": "actions.intent.MAIN",
            "locale": "fr-FR"
        }
    }
}

r = requests.request("POST", "https://actions.googleapis.com/v2/conversations:send", data=json.dumps(payload), headers=headers)
print(str(r.status_code) + ": " + r.text)



from jinja2 import Template, FileSystemLoader, Environment


class TemplatesAccess:
    file_loader = FileSystemLoader("templates")
    env = Environment(loader=file_loader)

    _endpoints_template = None

    @staticmethod
    def _get_template(template_name: str) -> Template:
        return TemplatesAccess.env.get_template(template_name)

    @property
    def endpoints_template(self) -> Template:
        if TemplatesAccess._endpoints_template is None:
            TemplatesAccess._endpoints_template = TemplatesAccess._get_template("endpoints.tem")
        return TemplatesAccess._endpoints_template



#!/usr/bin/env python3
"""Create a recording with arbitrary duration.

The soundfile module (https://PySoundFile.readthedocs.io/) has to be installed!

"""
import argparse
import os
import tempfile
import queue
import sys
import time

import sounddevice as sd
import soundfile as sf
import numpy  # Make sure NumPy is loaded before it is used in the callback
assert numpy  # avoid "imported but unused" message (W0611)


def int_or_str(text):
    """Helper function for argument parsing."""
    try:
        return int(text)
    except ValueError:
        return text


parser = argparse.ArgumentParser(add_help=False)
parser.add_argument(
    '-l', '--list-devices', action='store_true',
    help='show list of audio devices and exit')
args, remaining = parser.parse_known_args()
if args.list_devices:
    print(sd.query_devices())
    parser.exit(0)
parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    parents=[parser])
parser.add_argument(
    'filename', nargs='?', metavar='FILENAME',
    help='audio file to store recording to')
parser.add_argument(
    '-d', '--device', type=int_or_str,
    help='input device (numeric ID or substring)')
parser.add_argument(
    '-r', '--samplerate', type=int, help='sampling rate')
parser.add_argument(
    '-c', '--channels', type=int, default=1, help='number of input channels')
parser.add_argument(
    '-t', '--subtype', type=str, help='sound file subtype (e.g. "PCM_24")')
args = parser.parse_args(remaining)

q = queue.Queue()


def callback(indata, frames, time, status):
    """This is called (from a separate thread) for each audio block."""
    if status:
        print(status, file=sys.stderr)
    q.put(indata.copy())

class AudioRecording:
    def __init__(self):
        self.target_filepath = "/inoft_vocal_framework/platforms_handlers/c#/test.wav"
        # todo: add support for mp3
        self.samplerate = 44100
        self.num_channels = 2
        self.sound_file_subtype = "PCM_24"
        self.device_id = 1

    def start_recording(self):
        time_start = time.time()
        if args.samplerate is None:
            device_info = sd.query_devices(self.device_id, 'input')
            # soundfile expects an int, sounddevice provides a float:
            args.samplerate = int(device_info['default_samplerate'])
        if args.filename is None:
            args.filename = tempfile.mktemp(prefix='delme_rec_unlimited_',
                                            suffix='.wav', dir='')

        if os.path.exists(self.target_filepath):
            os.remove(self.target_filepath)

        # Make sure the file is opened before recording anything:
        with sf.SoundFile(self.target_filepath, mode="x", samplerate=self.samplerate,
                          channels=self.num_channels, subtype=self.sound_file_subtype) as file:
            with sd.InputStream(samplerate=self.samplerate, device=self.device_id,
                                channels=self.num_channels, callback=callback):
                print('#' * 80)
                print('press Ctrl+C to stop the recording')
                print('#' * 80)
                while True:
                    file.write(q.get())
                    if time.time() > time_start + 10:
                        self.end_recording()

    def end_recording(self):
        print(f"\nRecording finished: {repr(args.filename)}")
        parser.exit(0)


if __name__ == "__main__":
    AudioRecording().start_recording()

# except Exception as e:
#    parser.exit(type(e).__name__ + ': ' + str(e))




class ClientWrapper:
    def __init__(self, prefix: str = "."):
        from inoft_vocal_framework.platforms_handlers.discord.discord_static_infos import DiscordStaticInfos
        DiscordStaticInfos.COMMAND_PREFIX = prefix

        from inoft_vocal_framework.platforms_handlers.discord import backend_client
        self.client = backend_client.bot_client

    def run(self, token: str):
        self.client.run(token)

    async def connect(self, *, timeout=60.0, reconnect=True):
        """
        key_id, _ = self._get_voice_client_key()
        state = self._state

        if state._get_voice_client(key_id):
            raise ClientException('Already connected to a voice channel.')
        """

        from inoft_vocal_framework.platforms_handlers.discord.voice_client import VoiceClient
        voice = VoiceClient(state=state, timeout=timeout, channel=self)
        state._add_voice_client(key_id, voice)

        try:
            await voice.connect(reconnect=reconnect)
        except asyncio.TimeoutError:
            try:
                await voice.disconnect(force=True)
            except Exception:
                # we don't care if disconnect failed because connection failed
                pass
            raise # re-raise

        return voice


if __name__ == "__main__":
    from inoft_vocal_framework.platforms_handlers.discord.static_token import token
    client_wrapper = ClientWrapper(prefix="_")
    client_wrapper.run(token=token)

